# KCC topology utilities
#
# Copyright (C) Dave Craft 2011
# Copyright (C) Jelmer Vernooij 2011
# Copyright (C) Andrew Bartlett 2015
#
# Andrew Bartlett's alleged work performed by his underlings Douglas
# Bagnall and Garming Sam.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
import ldb
import uuid

from samba import dsdb
from samba.dcerpc import (
    drsblobs,
    drsuapi,
    misc,
)
from samba.samdb import dsdb_dn_guess, BinaryDn
from samba.ndr import ndr_unpack, ndr_pack
from collections import Counter


class KCCError(Exception):
    pass


class NCType(object):
    (unknown, schema, domain, config, application) = range(0, 5)


# map the NCType enum to strings for debugging
nctype_lut = dict((v, k) for k, v in NCType.__dict__.items() if k[:2] != '__')


class NamingContext(object):
    """Base class for a naming context.

    Holds the DN, GUID, SID (if available) and type of the DN.
    Subclasses may inherit from this and specialize
    """

    def __init__(self, nc_dnstr):
        """Instantiate a NamingContext

        :param nc_dnstr: NC dn string
        """
        self.nc_dnstr = nc_dnstr
        self.nc_guid = None
        self.nc_sid = None
        self.nc_type = NCType.unknown

    def __str__(self):
        """Debug dump string output of class"""
        text = "%s:" % (self.__class__.__name__,) +\
               "\n\tnc_dnstr=%s" % self.nc_dnstr +\
               "\n\tnc_guid=%s" % str(self.nc_guid)

        if self.nc_sid is None:
            text = text + "\n\tnc_sid=<absent>"
        else:
            text = text + "\n\tnc_sid=<present>"

        text = text + "\n\tnc_type=%s (%s)" % (nctype_lut[self.nc_type],
                                               self.nc_type)
        return text

    def load_nc(self, samdb):
        attrs = ["objectGUID",
                 "objectSid"]
        try:
            res = samdb.search(base=self.nc_dnstr,
                               scope=ldb.SCOPE_BASE, attrs=attrs)

        except ldb.LdbError as e:
            (enum, estr) = e.args
            raise KCCError("Unable to find naming context (%s) - (%s)" %
                           (self.nc_dnstr, estr))
        msg = res[0]
        if "objectGUID" in msg:
            self.nc_guid = misc.GUID(samdb.schema_format_value("objectGUID",
                                     msg["objectGUID"][0]))
        if "objectSid" in msg:
            self.nc_sid = msg["objectSid"][0]

        assert self.nc_guid is not None

    def is_config(self):
        """Return True if NC is config"""
        assert self.nc_type != NCType.unknown
        return self.nc_type == NCType.config

    def identify_by_basedn(self, samdb):
        """Given an NC object, identify what type it is thru
           the samdb basedn strings and NC sid value
        """
        # Invoke loader to initialize guid and more
        # importantly sid value (sid is used to identify
        # domain NCs)
        if self.nc_guid is None:
            self.load_nc(samdb)

        # We check against schema and config because they
        # will be the same for all nTDSDSAs in the forest.
        # That leaves the domain NCs which can be identified
        # by sid and application NCs as the last identified
        if self.nc_dnstr == str(samdb.get_schema_basedn()):
            self.nc_type = NCType.schema
        elif self.nc_dnstr == str(samdb.get_config_basedn()):
            self.nc_type = NCType.config
        elif self.nc_sid is not None:
            self.nc_type = NCType.domain
        else:
            self.nc_type = NCType.application

    def identify_by_dsa_attr(self, samdb, attr):
        """Given an NC which has been discovered thru the
        nTDSDSA database object, determine what type of NC
        it is (i.e. schema, config, domain, application) via
        the use of the schema attribute under which the NC
        was found.

        :param attr: attr of nTDSDSA object where NC DN appears
        """
        # If the NC is listed under msDS-HasDomainNCs then
        # this can only be a domain NC and it is our default
        # domain for this dsa
        if attr == "msDS-HasDomainNCs":
            self.nc_type = NCType.domain

        # If the NC is listed under hasPartialReplicaNCs
        # this is only a domain NC
        elif attr == "hasPartialReplicaNCs":
            self.nc_type = NCType.domain

        # NCs listed under hasMasterNCs are either
        # default domain, schema, or config.  We
        # utilize the identify_by_basedn() to
        # identify those
        elif attr == "hasMasterNCs":
            self.identify_by_basedn(samdb)

        # Still unknown (unlikely) but for completeness
        # and for finally identifying application NCs
        if self.nc_type == NCType.unknown:
            self.identify_by_basedn(samdb)


class NCReplica(NamingContext):
    """Naming context replica that is relative to a specific DSA.

    This is a more specific form of NamingContext class (inheriting from that
    class) and it identifies unique attributes of the DSA's replica for a NC.
    """

    def __init__(self, dsa, nc_dnstr):
        """Instantiate a Naming Context Replica

        :param dsa_guid: GUID of DSA where replica appears
        :param nc_dnstr: NC dn string
        """
        self.rep_dsa_dnstr = dsa.dsa_dnstr
        self.rep_dsa_guid = dsa.dsa_guid
        self.rep_default = False  # replica for DSA's default domain
        self.rep_partial = False
        self.rep_ro = False
        self.rep_instantiated_flags = 0

        self.rep_fsmo_role_owner = None

        # RepsFromTo tuples
        self.rep_repsFrom = []

        # RepsFromTo tuples
        self.rep_repsTo = []

        # The (is present) test is a combination of being
        # enumerated in (hasMasterNCs or msDS-hasFullReplicaNCs or
        # hasPartialReplicaNCs) as well as its replica flags found
        # thru the msDS-HasInstantiatedNCs.  If the NC replica meets
        # the first enumeration test then this flag is set true
        self.rep_present_criteria_one = False

        # Call my super class we inherited from
        NamingContext.__init__(self, nc_dnstr)

    def __str__(self):
        """Debug dump string output of class"""
        text = "%s:" % self.__class__.__name__ +\
               "\n\tdsa_dnstr=%s" % self.rep_dsa_dnstr +\
               "\n\tdsa_guid=%s" % self.rep_dsa_guid +\
               "\n\tdefault=%s" % self.rep_default +\
               "\n\tro=%s" % self.rep_ro +\
               "\n\tpartial=%s" % self.rep_partial +\
               "\n\tpresent=%s" % self.is_present() +\
               "\n\tfsmo_role_owner=%s" % self.rep_fsmo_role_owner +\
               "".join("\n%s" % rep for rep in self.rep_repsFrom) +\
               "".join("\n%s" % rep for rep in self.rep_repsTo)

        return "%s\n%s" % (NamingContext.__str__(self), text)

    def set_instantiated_flags(self, flags=0):
        """Set or clear NC replica instantiated flags"""
        self.rep_instantiated_flags = flags

    def identify_by_dsa_attr(self, samdb, attr):
        """Given an NC which has been discovered thru the
        nTDSDSA database object, determine what type of NC
        replica it is (i.e. partial, read only, default)

        :param attr: attr of nTDSDSA object where NC DN appears
        """
        # If the NC was found under hasPartialReplicaNCs
        # then a partial replica at this dsa
        if attr == "hasPartialReplicaNCs":
            self.rep_partial = True
            self.rep_present_criteria_one = True

        # If the NC is listed under msDS-HasDomainNCs then
        # this can only be a domain NC and it is the DSA's
        # default domain NC
        elif attr == "msDS-HasDomainNCs":
            self.rep_default = True

        # NCs listed under hasMasterNCs are either
        # default domain, schema, or config.  We check
        # against schema and config because they will be
        # the same for all nTDSDSAs in the forest.  That
        # leaves the default domain NC remaining which
        # may be different for each nTDSDSAs (and thus
        # we don't compare against this samdb's default
        # basedn
        elif attr == "hasMasterNCs":
            self.rep_present_criteria_one = True

            if self.nc_dnstr != str(samdb.get_schema_basedn()) and \
               self.nc_dnstr != str(samdb.get_config_basedn()):
                self.rep_default = True

        # RODC only
        elif attr == "msDS-hasFullReplicaNCs":
            self.rep_present_criteria_one = True
            self.rep_ro = True

        # Not RODC
        elif attr == "msDS-hasMasterNCs":
            self.rep_present_criteria_one = True
            self.rep_ro = False

        # Now use this DSA attribute to identify the naming
        # context type by calling the super class method
        # of the same name
        NamingContext.identify_by_dsa_attr(self, samdb, attr)

    def is_default(self):
        """Whether this is a default domain for the dsa that this NC appears on
        """
        return self.rep_default

    def is_ro(self):
        """Return True if NC replica is read only"""
        return self.rep_ro

    def is_partial(self):
        """Return True if NC replica is partial"""
        return self.rep_partial

    def is_present(self):
        """Given an NC replica which has been discovered thru the
        nTDSDSA database object and populated with replica flags
        from the msDS-HasInstantiatedNCs; return whether the NC
        replica is present (true) or if the IT_NC_GOING flag is
        set then the NC replica is not present (false)
        """
        if self.rep_present_criteria_one and \
           self.rep_instantiated_flags & dsdb.INSTANCE_TYPE_NC_GOING == 0:
            return True
        return False

    def load_repsFrom(self, samdb):
        """Given an NC replica which has been discovered thru the nTDSDSA
        database object, load the repsFrom attribute for the local replica.
        held by my dsa.  The repsFrom attribute is not replicated so this
        attribute is relative only to the local DSA that the samdb exists on
        """
        try:
            res = samdb.search(base=self.nc_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=["repsFrom"])

        except ldb.LdbError as e1:
            (enum, estr) = e1.args
            raise KCCError("Unable to find NC for (%s) - (%s)" %
                           (self.nc_dnstr, estr))

        msg = res[0]

        # Possibly no repsFrom if this is a singleton DC
        if "repsFrom" in msg:
            for value in msg["repsFrom"]:
                try:
                    unpacked = ndr_unpack(drsblobs.repsFromToBlob, value)
                except RuntimeError as e:
                    print("bad repsFrom NDR: %r" % (value),
                          file=sys.stderr)
                    continue
                rep = RepsFromTo(self.nc_dnstr, unpacked)
                self.rep_repsFrom.append(rep)

    def commit_repsFrom(self, samdb, ro=False):
        """Commit repsFrom to the database"""

        # XXX - This is not truly correct according to the MS-TECH
        #       docs.  To commit a repsFrom we should be using RPCs
        #       IDL_DRSReplicaAdd, IDL_DRSReplicaModify, and
        #       IDL_DRSReplicaDel to affect a repsFrom change.
        #
        #       Those RPCs are missing in samba, so I'll have to
        #       implement them to get this to more accurately
        #       reflect the reference docs.  As of right now this
        #       commit to the database will work as its what the
        #       older KCC also did
        modify = False
        newreps = []
        delreps = []

        for repsFrom in self.rep_repsFrom:

            # Leave out any to be deleted from
            # replacement list.  Build a list
            # of to be deleted reps which we will
            # remove from rep_repsFrom list below
            if repsFrom.to_be_deleted:
                delreps.append(repsFrom)
                modify = True
                continue

            if repsFrom.is_modified():
                repsFrom.set_unmodified()
                modify = True

            # current (unmodified) elements also get
            # appended here but no changes will occur
            # unless something is "to be modified" or
            # "to be deleted"
            newreps.append(ndr_pack(repsFrom.ndr_blob))

        # Now delete these from our list of rep_repsFrom
        for repsFrom in delreps:
            self.rep_repsFrom.remove(repsFrom)
        delreps = []

        # Nothing to do if no reps have been modified or
        # need to be deleted or input option has informed
        # us to be "readonly" (ro).  Leave database
        # record "as is"
        if not modify or ro:
            return

        m = ldb.Message()
        m.dn = ldb.Dn(samdb, self.nc_dnstr)

        m["repsFrom"] = \
            ldb.MessageElement(newreps, ldb.FLAG_MOD_REPLACE, "repsFrom")

        try:
            samdb.modify(m)

        except ldb.LdbError as estr:
            raise KCCError("Could not set repsFrom for (%s) - (%s)" %
                           (self.nc_dnstr, estr))

    def load_replUpToDateVector(self, samdb):
        """Given an NC replica which has been discovered thru the nTDSDSA
        database object, load the replUpToDateVector attribute for the
        local replica. held by my dsa. The replUpToDateVector
        attribute is not replicated so this attribute is relative only
        to the local DSA that the samdb exists on

        """
        try:
            res = samdb.search(base=self.nc_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=["replUpToDateVector"])

        except ldb.LdbError as e2:
            (enum, estr) = e2.args
            raise KCCError("Unable to find NC for (%s) - (%s)" %
                           (self.nc_dnstr, estr))

        msg = res[0]

        # Possibly no replUpToDateVector if this is a singleton DC
        if "replUpToDateVector" in msg:
            value = msg["replUpToDateVector"][0]
            blob = ndr_unpack(drsblobs.replUpToDateVectorBlob,
                              value)
            if blob.version != 2:
                # Samba only generates version 2, and this runs locally
                raise AttributeError("Unexpected replUpToDateVector version %d"
                                     % blob.version)

            self.rep_replUpToDateVector_cursors = blob.ctr.cursors
        else:
            self.rep_replUpToDateVector_cursors = []

    def dumpstr_to_be_deleted(self):
        return '\n'.join(str(x) for x in self.rep_repsFrom if x.to_be_deleted)

    def dumpstr_to_be_modified(self):
        return '\n'.join(str(x) for x in self.rep_repsFrom if x.is_modified())

    def load_fsmo_roles(self, samdb):
        """Given an NC replica which has been discovered thru the nTDSDSA
        database object, load the fSMORoleOwner attribute.
        """
        try:
            res = samdb.search(base=self.nc_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=["fSMORoleOwner"])

        except ldb.LdbError as e3:
            (enum, estr) = e3.args
            raise KCCError("Unable to find NC for (%s) - (%s)" %
                           (self.nc_dnstr, estr))

        msg = res[0]

        # Possibly no fSMORoleOwner
        if "fSMORoleOwner" in msg:
            self.rep_fsmo_role_owner = msg["fSMORoleOwner"]

    def is_fsmo_role_owner(self, dsa_dnstr):
        if self.rep_fsmo_role_owner is not None and \
           self.rep_fsmo_role_owner == dsa_dnstr:
            return True
        return False

    def load_repsTo(self, samdb):
        """Given an NC replica which has been discovered thru the nTDSDSA
        database object, load the repsTo attribute for the local replica.
        held by my dsa.  The repsTo attribute is not replicated so this
        attribute is relative only to the local DSA that the samdb exists on

        This is responsible for push replication, not scheduled pull
        replication. Not to be confused for repsFrom.
        """
        try:
            res = samdb.search(base=self.nc_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=["repsTo"])

        except ldb.LdbError as e4:
            (enum, estr) = e4.args
            raise KCCError("Unable to find NC for (%s) - (%s)" %
                           (self.nc_dnstr, estr))

        msg = res[0]

        # Possibly no repsTo if this is a singleton DC
        if "repsTo" in msg:
            for value in msg["repsTo"]:
                try:
                    unpacked = ndr_unpack(drsblobs.repsFromToBlob, value)
                except RuntimeError as e:
                    print("bad repsTo NDR: %r" % (value),
                          file=sys.stderr)
                    continue
                rep = RepsFromTo(self.nc_dnstr, unpacked)
                self.rep_repsTo.append(rep)

    def commit_repsTo(self, samdb, ro=False):
        """Commit repsTo to the database"""

        # XXX - This is not truly correct according to the MS-TECH
        #       docs.  To commit a repsTo we should be using RPCs
        #       IDL_DRSReplicaAdd, IDL_DRSReplicaModify, and
        #       IDL_DRSReplicaDel to affect a repsTo change.
        #
        #       Those RPCs are missing in samba, so I'll have to
        #       implement them to get this to more accurately
        #       reflect the reference docs.  As of right now this
        #       commit to the database will work as its what the
        #       older KCC also did
        modify = False
        newreps = []
        delreps = []

        for repsTo in self.rep_repsTo:

            # Leave out any to be deleted from
            # replacement list.  Build a list
            # of to be deleted reps which we will
            # remove from rep_repsTo list below
            if repsTo.to_be_deleted:
                delreps.append(repsTo)
                modify = True
                continue

            if repsTo.is_modified():
                repsTo.set_unmodified()
                modify = True

            # current (unmodified) elements also get
            # appended here but no changes will occur
            # unless something is "to be modified" or
            # "to be deleted"
            newreps.append(ndr_pack(repsTo.ndr_blob))

        # Now delete these from our list of rep_repsTo
        for repsTo in delreps:
            self.rep_repsTo.remove(repsTo)
        delreps = []

        # Nothing to do if no reps have been modified or
        # need to be deleted or input option has informed
        # us to be "readonly" (ro).  Leave database
        # record "as is"
        if not modify or ro:
            return

        m = ldb.Message()
        m.dn = ldb.Dn(samdb, self.nc_dnstr)

        m["repsTo"] = \
            ldb.MessageElement(newreps, ldb.FLAG_MOD_REPLACE, "repsTo")

        try:
            samdb.modify(m)

        except ldb.LdbError as estr:
            raise KCCError("Could not set repsTo for (%s) - (%s)" %
                           (self.nc_dnstr, estr))


class DirectoryServiceAgent(object):

    def __init__(self, dsa_dnstr):
        """Initialize DSA class.

        Class is subsequently fully populated by calling the load_dsa() method

        :param dsa_dnstr:  DN of the nTDSDSA
        """
        self.dsa_dnstr = dsa_dnstr
        self.dsa_guid = None
        self.dsa_ivid = None
        self.dsa_is_ro = False
        self.dsa_is_istg = False
        self.options = 0
        self.dsa_behavior = 0
        self.default_dnstr = None  # default domain dn string for dsa

        # NCReplicas for this dsa that are "present"
        # Indexed by DN string of naming context
        self.current_rep_table = {}

        # NCReplicas for this dsa that "should be present"
        # Indexed by DN string of naming context
        self.needed_rep_table = {}

        # NTDSConnections for this dsa.  These are current
        # valid connections that are committed or pending a commit
        # in the database.  Indexed by DN string of connection
        self.connect_table = {}

    def __str__(self):
        """Debug dump string output of class"""

        text = "%s:" % self.__class__.__name__
        if self.dsa_dnstr is not None:
            text = text + "\n\tdsa_dnstr=%s" % self.dsa_dnstr
        if self.dsa_guid is not None:
            text = text + "\n\tdsa_guid=%s" % str(self.dsa_guid)
        if self.dsa_ivid is not None:
            text = text + "\n\tdsa_ivid=%s" % str(self.dsa_ivid)

        text += "\n\tro=%s" % self.is_ro() +\
                "\n\tgc=%s" % self.is_gc() +\
                "\n\tistg=%s" % self.is_istg() +\
                "\ncurrent_replica_table:" +\
                "\n%s" % self.dumpstr_current_replica_table() +\
                "\nneeded_replica_table:" +\
                "\n%s" % self.dumpstr_needed_replica_table() +\
                "\nconnect_table:" +\
                "\n%s" % self.dumpstr_connect_table()

        return text

    def get_current_replica(self, nc_dnstr):
        return self.current_rep_table.get(nc_dnstr)

    def is_istg(self):
        """Returns True if dsa is intersite topology generator for it's site"""
        # The KCC on an RODC always acts as an ISTG for itself
        return self.dsa_is_istg or self.dsa_is_ro

    def is_ro(self):
        """Returns True if dsa a read only domain controller"""
        return self.dsa_is_ro

    def is_gc(self):
        """Returns True if dsa hosts a global catalog"""
        if (self.options & dsdb.DS_NTDSDSA_OPT_IS_GC) != 0:
            return True
        return False

    def is_minimum_behavior(self, version):
        """Is dsa at minimum windows level greater than or equal to (version)

        :param version: Windows version to test against
            (e.g. DS_DOMAIN_FUNCTION_2008)
        """
        if self.dsa_behavior >= version:
            return True
        return False

    def is_translate_ntdsconn_disabled(self):
        """Whether this allows NTDSConnection translation in its options."""
        if (self.options & dsdb.DS_NTDSDSA_OPT_DISABLE_NTDSCONN_XLATE) != 0:
            return True
        return False

    def get_rep_tables(self):
        """Return DSA current and needed replica tables
        """
        return self.current_rep_table, self.needed_rep_table

    def get_parent_dnstr(self):
        """Get the parent DN string of this object."""
        head, sep, tail = self.dsa_dnstr.partition(',')
        return tail

    def load_dsa(self, samdb):
        """Load a DSA from the samdb.

        Prior initialization has given us the DN of the DSA that we are to
        load.  This method initializes all other attributes, including loading
        the NC replica table for this DSA.
        """
        attrs = ["objectGUID",
                 "invocationID",
                 "options",
                 "msDS-isRODC",
                 "msDS-Behavior-Version"]
        try:
            res = samdb.search(base=self.dsa_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=attrs)

        except ldb.LdbError as e5:
            (enum, estr) = e5.args
            raise KCCError("Unable to find nTDSDSA for (%s) - (%s)" %
                           (self.dsa_dnstr, estr))

        msg = res[0]
        self.dsa_guid = misc.GUID(samdb.schema_format_value("objectGUID",
                                  msg["objectGUID"][0]))

        # RODCs don't originate changes and thus have no invocationId,
        # therefore we must check for existence first
        if "invocationId" in msg:
            self.dsa_ivid = misc.GUID(samdb.schema_format_value("objectGUID",
                                      msg["invocationId"][0]))

        if "options" in msg:
            self.options = int(msg["options"][0])

        if "msDS-isRODC" in msg and str(msg["msDS-isRODC"][0]) == "TRUE":
            self.dsa_is_ro = True
        else:
            self.dsa_is_ro = False

        if "msDS-Behavior-Version" in msg:
            self.dsa_behavior = int(msg['msDS-Behavior-Version'][0])

        # Load the NC replicas that are enumerated on this dsa
        self.load_current_replica_table(samdb)

        # Load the nTDSConnection that are enumerated on this dsa
        self.load_connection_table(samdb)

    def load_current_replica_table(self, samdb):
        """Method to load the NC replica's listed for DSA object.

        This method queries the samdb for (hasMasterNCs, msDS-hasMasterNCs,
        hasPartialReplicaNCs, msDS-HasDomainNCs, msDS-hasFullReplicaNCs, and
        msDS-HasInstantiatedNCs) to determine complete list of NC replicas that
        are enumerated for the DSA.  Once a NC replica is loaded it is
        identified (schema, config, etc) and the other replica attributes
        (partial, ro, etc) are determined.

        :param samdb: database to query for DSA replica list
        """
        ncattrs = [
            # not RODC - default, config, schema (old style)
            "hasMasterNCs",
            # not RODC - default, config, schema, app NCs
            "msDS-hasMasterNCs",
            # domain NC partial replicas
            "hasPartialReplicaNCs",
            # default domain NC
            "msDS-HasDomainNCs",
            # RODC only - default, config, schema, app NCs
            "msDS-hasFullReplicaNCs",
            # Identifies if replica is coming, going, or stable
            "msDS-HasInstantiatedNCs"
        ]
        try:
            res = samdb.search(base=self.dsa_dnstr, scope=ldb.SCOPE_BASE,
                               attrs=ncattrs)

        except ldb.LdbError as e6:
            (enum, estr) = e6.args
            raise KCCError("Unable to find nTDSDSA NCs for (%s) - (%s)" %
                           (self.dsa_dnstr, estr))

        # The table of NCs for the dsa we are searching
        tmp_table = {}

        # We should get one response to our query here for
        # the ntds that we requested
        if len(res[0]) > 0:

            # Our response will contain a number of elements including
            # the dn of the dsa as well as elements for each
            # attribute (e.g. hasMasterNCs).  Each of these elements
            # is a dictionary list which we retrieve the keys for and
            # then iterate over them
            for k in res[0].keys():
                if k == "dn":
                    continue

                # For each attribute type there will be one or more DNs
                # listed.  For instance DCs normally have 3 hasMasterNCs
                # listed.
                for value in res[0][k]:
                    # msDS-HasInstantiatedNCs is a BinaryDN, but the
                    # others are plain DNs.
                    dsdn = dsdb_dn_guess(samdb, value)
                    dnstr = str(dsdn.dn)

                    if dnstr not in tmp_table:
                        rep = NCReplica(self, dnstr)
                        tmp_table[dnstr] = rep
                    else:
                        rep = tmp_table[dnstr]

                    if k == "msDS-HasInstantiatedNCs":
                        # msDS-HasInstantiatedNCs should only be DN+Binary
                        # (MS-ADTS 6.1.1.2.2.1.2.1.1 and 6.1.2.3.1)
                        # but sometimes we see it as a plain DN.
                        if isinstance(dsdn, BinaryDn):
                            flags = dsdn.get_binary_integer()
                            rep.set_instantiated_flags(flags)
                        else:
                            print("msDS-HasInstantiatedNCsis not a BinaryDn '{dsdn}'",
                                  file=sys.stderr)
                        continue

                    rep.identify_by_dsa_attr(samdb, k)

                    # if we've identified the default domain NC
                    # then save its DN string
                    if rep.is_default():
                        self.default_dnstr = dnstr
        else:
            raise KCCError("No nTDSDSA NCs for (%s)" % self.dsa_dnstr)

        # Assign our newly built NC replica table to this dsa
        self.current_rep_table = tmp_table

    def add_needed_replica(self, rep):
        """Method to add a NC replica that "should be present" to the
        needed_rep_table.
        """
        self.needed_rep_table[rep.nc_dnstr] = rep

    def load_connection_table(self, samdb):
        """Method to load the nTDSConnections listed for DSA object.

        :param samdb: database to query for DSA connection list
        """
        try:
            res = samdb.search(base=self.dsa_dnstr,
                               scope=ldb.SCOPE_SUBTREE,
                               expression="(objectClass=nTDSConnection)")

        except ldb.LdbError as e7:
            (enum, estr) = e7.args
            raise KCCError("Unable to find nTDSConnection for (%s) - (%s)" %
                           (self.dsa_dnstr, estr))

        for msg in res:
            dnstr = str(msg.dn)

            # already loaded
            if dnstr in self.connect_table:
                continue

            connect = NTDSConnection(dnstr)

            connect.load_connection(samdb)
            self.connect_table[dnstr] = connect

    def commit_connections(self, samdb, ro=False):
        """Method to commit any uncommitted nTDSConnections
        modifications that are in our table.  These would be
        identified connections that are marked to be added or
        deleted

        :param samdb: database to commit DSA connection list to
        :param ro: if (true) then perform internal operations but
            do not write to the database (readonly)
        """
        delconn = []

        for dnstr, connect in self.connect_table.items():
            if connect.to_be_added:
                connect.commit_added(samdb, ro)

            if connect.to_be_modified:
                connect.commit_modified(samdb, ro)

            if connect.to_be_deleted:
                connect.commit_deleted(samdb, ro)
                delconn.append(dnstr)

        # Now delete the connection from the table
        for dnstr in delconn:
            del self.connect_table[dnstr]

    def add_connection(self, dnstr, connect):
        assert dnstr not in self.connect_table
        self.connect_table[dnstr] = connect

    def get_connection_by_from_dnstr(self, from_dnstr):
        """Scan DSA nTDSConnection table and return connection
        with a "fromServer" dn string equivalent to method
        input parameter.

        :param from_dnstr: search for this from server entry
        """
        answer = []
        for connect in self.connect_table.values():
            if connect.get_from_dnstr() == from_dnstr:
                answer.append(connect)

        return answer

    def dumpstr_current_replica_table(self):
        """Debug dump string output of current replica table"""
        return '\n'.join(str(x) for x in self.current_rep_table)

    def dumpstr_needed_replica_table(self):
        """Debug dump string output of needed replica table"""
        return '\n'.join(str(x) for x in self.needed_rep_table)

    def dumpstr_connect_table(self):
        """Debug dump string output of connect table"""
        return '\n'.join(str(x) for x in self.connect_table)

    def new_connection(self, options, system_flags, transport, from_dnstr,
                       sched):
        """Set up a new connection for the DSA based on input
        parameters.  Connection will be added to the DSA
        connect_table and will be marked as "to be added" pending
        a call to commit_connections()
        """
        dnstr = "CN=%s," % str(uuid.uuid4()) + self.dsa_dnstr

        connect = NTDSConnection(dnstr)
        connect.to_be_added = True
        connect.enabled = True
        connect.from_dnstr = from_dnstr
        connect.options = options
        connect.system_flags = system_flags

        if transport is not None:
            connect.transport_dnstr = transport.dnstr
            connect.transport_guid = transport.guid

        if sched is not None:
            connect.schedule = sched
        else:
            # Create schedule.  Attribute value set according to MS-TECH
            # intra-site connection creation document
            connect.schedule = new_connection_schedule()

        self.add_connection(dnstr, connect)
        return connect


class NTDSConnection(object):
    """Class defines a nTDSConnection found under a DSA
    """
    def __init__(self, dnstr):
        self.dnstr = dnstr
        self.guid = None
        self.enabled = False
        self.whenCreated = 0
        self.to_be_added = False  # new connection needs to be added
        self.to_be_deleted = False  # old connection needs to be deleted
        self.to_be_modified = False
        self.options = 0
        self.system_flags = 0
        self.transport_dnstr = None
        self.transport_guid = None
        self.from_dnstr = None
        self.schedule = None

    def __str__(self):
        """Debug dump string output of NTDSConnection object"""

        text = "%s:\n\tdn=%s" % (self.__class__.__name__, self.dnstr) +\
               "\n\tenabled=%s" % self.enabled +\
               "\n\tto_be_added=%s" % self.to_be_added +\
               "\n\tto_be_deleted=%s" % self.to_be_deleted +\
               "\n\tto_be_modified=%s" % self.to_be_modified +\
               "\n\toptions=0x%08X" % self.options +\
               "\n\tsystem_flags=0x%08X" % self.system_flags +\
               "\n\twhenCreated=%d" % self.whenCreated +\
               "\n\ttransport_dn=%s" % self.transport_dnstr

        if self.guid is not None:
            text += "\n\tguid=%s" % str(self.guid)

        if self.transport_guid is not None:
            text += "\n\ttransport_guid=%s" % str(self.transport_guid)

        text = text + "\n\tfrom_dn=%s" % self.from_dnstr

        if self.schedule is not None:
            text += "\n\tschedule.size=%s" % self.schedule.size +\
                    "\n\tschedule.bandwidth=%s" % self.schedule.bandwidth +\
                    ("\n\tschedule.numberOfSchedules=%s" %
                     self.schedule.numberOfSchedules)

            for i, header in enumerate(self.schedule.headerArray):
                text += ("\n\tschedule.headerArray[%d].type=%d" %
                        (i, header.type)) +\
                        ("\n\tschedule.headerArray[%d].offset=%d" %
                         (i, header.offset)) +\
                        "\n\tschedule.dataArray[%d].slots[ " % i +\
                        "".join("0x%X " % slot for slot in self.schedule.dataArray[i].slots) +\
                        "]"

        return text

    def load_connection(self, samdb):
        """Given a NTDSConnection object with an prior initialization
        for the object's DN, search for the DN and load attributes
        from the samdb.
        """
        attrs = ["options",
                 "enabledConnection",
                 "schedule",
                 "whenCreated",
                 "objectGUID",
                 "transportType",
                 "fromServer",
                 "systemFlags"]
        try:
            res = samdb.search(base=self.dnstr, scope=ldb.SCOPE_BASE,
                               attrs=attrs)

        except ldb.LdbError as e8:
            (enum, estr) = e8.args
            raise KCCError("Unable to find nTDSConnection for (%s) - (%s)" %
                           (self.dnstr, estr))

        msg = res[0]

        if "options" in msg:
            self.options = int(msg["options"][0])

        if "enabledConnection" in msg:
            if str(msg["enabledConnection"][0]).upper().lstrip().rstrip() == "TRUE":
                self.enabled = True

        if "systemFlags" in msg:
            self.system_flags = int(msg["systemFlags"][0])

        try:
            self.guid = \
                misc.GUID(samdb.schema_format_value("objectGUID",
                                                    msg["objectGUID"][0]))
        except KeyError:
            raise KCCError("Unable to find objectGUID in nTDSConnection "
                           "for (%s)" % (self.dnstr))

        if "transportType" in msg:
            dsdn = dsdb_dn_guess(samdb, msg["transportType"][0])
            self.load_connection_transport(samdb, str(dsdn.dn))

        if "schedule" in msg:
            self.schedule = ndr_unpack(drsblobs.schedule, msg["schedule"][0])

        if "whenCreated" in msg:
            self.whenCreated = ldb.string_to_time(str(msg["whenCreated"][0]))

        if "fromServer" in msg:
            dsdn = dsdb_dn_guess(samdb, msg["fromServer"][0])
            self.from_dnstr = str(dsdn.dn)
            assert self.from_dnstr is not None

    def load_connection_transport(self, samdb, tdnstr):
        """Given a NTDSConnection object which enumerates a transport
        DN, load the transport information for the connection object

        :param tdnstr: transport DN to load
        """
        attrs = ["objectGUID"]
        try:
            res = samdb.search(base=tdnstr,
                               scope=ldb.SCOPE_BASE, attrs=attrs)

        except ldb.LdbError as e9:
            (enum, estr) = e9.args
            raise KCCError("Unable to find transport (%s) - (%s)" %
                           (tdnstr, estr))

        if "objectGUID" in res[0]:
            msg = res[0]
            self.transport_dnstr = tdnstr
            self.transport_guid = \
                misc.GUID(samdb.schema_format_value("objectGUID",
                                                    msg["objectGUID"][0]))
        assert self.transport_dnstr is not None
        assert self.transport_guid is not None

    def commit_deleted(self, samdb, ro=False):
        """Local helper routine for commit_connections() which
        handles committed connections that are to be deleted from
        the database database
        """
        assert self.to_be_deleted
        self.to_be_deleted = False

        # No database modification requested
        if ro:
            return

        try:
            samdb.delete(self.dnstr)
        except ldb.LdbError as e10:
            (enum, estr) = e10.args
            raise KCCError("Could not delete nTDSConnection for (%s) - (%s)" %
                           (self.dnstr, estr))

    def commit_added(self, samdb, ro=False):
        """Local helper routine for commit_connections() which
        handles committed connections that are to be added to the
        database
        """
        assert self.to_be_added
        self.to_be_added = False

        # No database modification requested
        if ro:
            return

        # First verify we don't have this entry to ensure nothing
        # is programmatically amiss
        found = False
        try:
            msg = samdb.search(base=self.dnstr, scope=ldb.SCOPE_BASE)
            if len(msg) != 0:
                found = True

        except ldb.LdbError as e11:
            (enum, estr) = e11.args
            if enum != ldb.ERR_NO_SUCH_OBJECT:
                raise KCCError("Unable to search for (%s) - (%s)" %
                               (self.dnstr, estr))
        if found:
            raise KCCError("nTDSConnection for (%s) already exists!" %
                           self.dnstr)

        if self.enabled:
            enablestr = "TRUE"
        else:
            enablestr = "FALSE"

        # Prepare a message for adding to the samdb
        m = ldb.Message()
        m.dn = ldb.Dn(samdb, self.dnstr)

        m["objectClass"] = \
            ldb.MessageElement("nTDSConnection", ldb.FLAG_MOD_ADD,
                               "objectClass")
        m["showInAdvancedViewOnly"] = \
            ldb.MessageElement("TRUE", ldb.FLAG_MOD_ADD,
                               "showInAdvancedViewOnly")
        m["enabledConnection"] = \
            ldb.MessageElement(enablestr, ldb.FLAG_MOD_ADD,
                               "enabledConnection")
        m["fromServer"] = \
            ldb.MessageElement(self.from_dnstr, ldb.FLAG_MOD_ADD, "fromServer")
        m["options"] = \
            ldb.MessageElement(str(self.options), ldb.FLAG_MOD_ADD, "options")
        m["systemFlags"] = \
            ldb.MessageElement(str(self.system_flags), ldb.FLAG_MOD_ADD,
                               "systemFlags")

        if self.transport_dnstr is not None:
            m["transportType"] = \
                ldb.MessageElement(str(self.transport_dnstr), ldb.FLAG_MOD_ADD,
                                   "transportType")

        if self.schedule is not None:
            m["schedule"] = \
                ldb.MessageElement(ndr_pack(self.schedule),
                                   ldb.FLAG_MOD_ADD, "schedule")
        try:
            samdb.add(m)
        except ldb.LdbError as e12:
            (enum, estr) = e12.args
            raise KCCError("Could not add nTDSConnection for (%s) - (%s)" %
                           (self.dnstr, estr))

    def commit_modified(self, samdb, ro=False):
        """Local helper routine for commit_connections() which
        handles committed connections that are to be modified to the
        database
        """
        assert self.to_be_modified
        self.to_be_modified = False

        # No database modification requested
        if ro:
            return

        # First verify we have this entry to ensure nothing
        # is programmatically amiss
        try:
            # we don't use the search result, but it tests the status
            # of self.dnstr in the database.
            samdb.search(base=self.dnstr, scope=ldb.SCOPE_BASE)

        except ldb.LdbError as e13:
            (enum, estr) = e13.args
            if enum == ldb.ERR_NO_SUCH_OBJECT:
                raise KCCError("nTDSConnection for (%s) doesn't exist!" %
                               self.dnstr)
            raise KCCError("Unable to search for (%s) - (%s)" %
                           (self.dnstr, estr))

        if self.enabled:
            enablestr = "TRUE"
        else:
            enablestr = "FALSE"

        # Prepare a message for modifying the samdb
        m = ldb.Message()
        m.dn = ldb.Dn(samdb, self.dnstr)

        m["enabledConnection"] = \
            ldb.MessageElement(enablestr, ldb.FLAG_MOD_REPLACE,
                               "enabledConnection")
        m["fromServer"] = \
            ldb.MessageElement(self.from_dnstr, ldb.FLAG_MOD_REPLACE,
                               "fromServer")
        m["options"] = \
            ldb.MessageElement(str(self.options), ldb.FLAG_MOD_REPLACE,
                               "options")
        m["systemFlags"] = \
            ldb.MessageElement(str(self.system_flags), ldb.FLAG_MOD_REPLACE,
                               "systemFlags")

        if self.transport_dnstr is not None:
            m["transportType"] = \
                ldb.MessageElement(str(self.transport_dnstr),
                                   ldb.FLAG_MOD_REPLACE, "transportType")
        else:
            m["transportType"] = \
                ldb.MessageElement([], ldb.FLAG_MOD_DELETE, "transportType")

        if self.schedule is not None:
            m["schedule"] = \
                ldb.MessageElement(ndr_pack(self.schedule),
                                   ldb.FLAG_MOD_REPLACE, "schedule")
        else:
            m["schedule"] = \
                ldb.MessageElement([], ldb.FLAG_MOD_DELETE, "schedule")
        try:
            samdb.modify(m)
        except ldb.LdbError as e14:
            (enum, estr) = e14.args
            raise KCCError("Could not modify nTDSConnection for (%s) - (%s)" %
                           (self.dnstr, estr))

    def set_modified(self, truefalse):
        self.to_be_modified = truefalse

    def is_schedule_minimum_once_per_week(self):
        """Returns True if our schedule includes at least one
        replication interval within the week.  False otherwise
        """
        # replinfo schedule is None means "always", while
        # NTDSConnection schedule is None means "never".
        if self.schedule is None or self.schedule.dataArray[0] is None:
            return False

        for slot in self.schedule.dataArray[0].slots:
            if (slot & 0x0F) != 0x0:
                return True
        return False

    def is_equivalent_schedule(self, sched):
        """Returns True if our schedule is equivalent to the input
        comparison schedule.

        :param shed: schedule to compare to
        """
        # There are 4 cases, where either self.schedule or sched can be None
        #
        #                   |  self. is None  |   self. is not None
        #     --------------+-----------------+--------------------
        #     sched is None |     True        |     False
        #     --------------+-----------------+--------------------
        # sched is not None |    False        |    do calculations

        if self.schedule is None:
            return sched is None

        if sched is None:
            return False

        if ((self.schedule.size != sched.size or
             self.schedule.bandwidth != sched.bandwidth or
             self.schedule.numberOfSchedules != sched.numberOfSchedules)):
            return False

        for i, header in enumerate(self.schedule.headerArray):

            if self.schedule.headerArray[i].type != sched.headerArray[i].type:
                return False

            if self.schedule.headerArray[i].offset != \
               sched.headerArray[i].offset:
                return False

            for a, b in zip(self.schedule.dataArray[i].slots,
                            sched.dataArray[i].slots):
                if a != b:
                    return False
        return True

    def is_rodc_topology(self):
        """Returns True if NTDS Connection specifies RODC
        topology only
        """
        if self.options & dsdb.NTDSCONN_OPT_RODC_TOPOLOGY == 0:
            return False
        return True

    def is_generated(self):
        """Returns True if NTDS Connection was generated by the
        KCC topology algorithm as opposed to set by the administrator
        """
        if self.options & dsdb.NTDSCONN_OPT_IS_GENERATED == 0:
            return False
        return True

    def is_override_notify_default(self):
        """Returns True if NTDS Connection should override notify default
        """
        if self.options & dsdb.NTDSCONN_OPT_OVERRIDE_NOTIFY_DEFAULT == 0:
            return False
        return True

    def is_use_notify(self):
        """Returns True if NTDS Connection should use notify
        """
        if self.options & dsdb.NTDSCONN_OPT_USE_NOTIFY == 0:
            return False
        return True

    def is_twoway_sync(self):
        """Returns True if NTDS Connection should use twoway sync
        """
        if self.options & dsdb.NTDSCONN_OPT_TWOWAY_SYNC == 0:
            return False
        return True

    def is_intersite_compression_disabled(self):
        """Returns True if NTDS Connection intersite compression
        is disabled
        """
        if self.options & dsdb.NTDSCONN_OPT_DISABLE_INTERSITE_COMPRESSION == 0:
            return False
        return True

    def is_user_owned_schedule(self):
        """Returns True if NTDS Connection has a user owned schedule
        """
        if self.options & dsdb.NTDSCONN_OPT_USER_OWNED_SCHEDULE == 0:
            return False
        return True

    def is_enabled(self):
        """Returns True if NTDS Connection is enabled
        """
        return self.enabled

    def get_from_dnstr(self):
        """Return fromServer dn string attribute"""
        return self.from_dnstr


class Partition(NamingContext):
    """A naming context discovered thru Partitions DN of the config schema.

    This is a more specific form of NamingContext class (inheriting from that
    class) and it identifies unique attributes enumerated in the Partitions
    such as which nTDSDSAs are cross referenced for replicas
    """
    def __init__(self, partstr):
        self.partstr = partstr
        self.enabled = True
        self.system_flags = 0
        self.rw_location_list = []
        self.ro_location_list = []

        # We don't have enough info to properly
        # fill in the naming context yet.  We'll get that
        # fully set up with load_partition().
        NamingContext.__init__(self, None)

    def load_partition(self, samdb):
        """Given a Partition class object that has been initialized with its
        partition dn string, load the partition from the sam database, identify
        the type of the partition (schema, domain, etc) and record the list of
        nTDSDSAs that appear in the cross reference attributes
        msDS-NC-Replica-Locations and msDS-NC-RO-Replica-Locations.

        :param samdb: sam database to load partition from
        """
        attrs = ["nCName",
                 "Enabled",
                 "systemFlags",
                 "msDS-NC-Replica-Locations",
                 "msDS-NC-RO-Replica-Locations"]
        try:
            res = samdb.search(base=self.partstr, scope=ldb.SCOPE_BASE,
                               attrs=attrs)

        except ldb.LdbError as e15:
            (enum, estr) = e15.args
            raise KCCError("Unable to find partition for (%s) - (%s)" %
                           (self.partstr, estr))
        msg = res[0]
        for k in msg.keys():
            if k == "dn":
                continue

            if k == "Enabled":
                if str(msg[k][0]).upper().lstrip().rstrip() == "TRUE":
                    self.enabled = True
                else:
                    self.enabled = False
                continue

            if k == "systemFlags":
                self.system_flags = int(msg[k][0])
                continue

            for value in msg[k]:
                dsdn = dsdb_dn_guess(samdb, value)
                dnstr = str(dsdn.dn)

                if k == "nCName":
                    self.nc_dnstr = dnstr
                    continue

                if k == "msDS-NC-Replica-Locations":
                    self.rw_location_list.append(dnstr)
                    continue

                if k == "msDS-NC-RO-Replica-Locations":
                    self.ro_location_list.append(dnstr)
                    continue

        # Now identify what type of NC this partition
        # enumerated
        self.identify_by_basedn(samdb)

    def is_enabled(self):
        """Returns True if partition is enabled
        """
        return self.is_enabled

    def is_foreign(self):
        """Returns True if this is not an Active Directory NC in our
        forest but is instead something else (e.g. a foreign NC)
        """
        if (self.system_flags & dsdb.SYSTEM_FLAG_CR_NTDS_NC) == 0:
            return True
        else:
            return False

    def should_be_present(self, target_dsa):
        """Tests whether this partition should have an NC replica
        on the target dsa.  This method returns a tuple of
        needed=True/False, ro=True/False, partial=True/False

        :param target_dsa: should NC be present on target dsa
        """
        ro = False
        partial = False

        # If this is the config, schema, or default
        # domain NC for the target dsa then it should
        # be present
        needed = (self.nc_type == NCType.config or
                  self.nc_type == NCType.schema or
                  (self.nc_type == NCType.domain and
                   self.nc_dnstr == target_dsa.default_dnstr))

        # A writable replica of an application NC should be present
        # if there a cross reference to the target DSA exists.  Depending
        # on whether the DSA is ro we examine which type of cross reference
        # to look for (msDS-NC-Replica-Locations or
        # msDS-NC-RO-Replica-Locations
        if self.nc_type == NCType.application:
            if target_dsa.is_ro():
                if target_dsa.dsa_dnstr in self.ro_location_list:
                    needed = True
            else:
                if target_dsa.dsa_dnstr in self.rw_location_list:
                    needed = True

        # If the target dsa is a gc then a partial replica of a
        # domain NC (other than the DSAs default domain) should exist
        # if there is also a cross reference for the DSA
        if (target_dsa.is_gc() and
            self.nc_type == NCType.domain and
            self.nc_dnstr != target_dsa.default_dnstr and
            (target_dsa.dsa_dnstr in self.ro_location_list or
             target_dsa.dsa_dnstr in self.rw_location_list)):
            needed = True
            partial = True

        # partial NCs are always readonly
        if needed and (target_dsa.is_ro() or partial):
            ro = True

        return needed, ro, partial

    def __str__(self):
        """Debug dump string output of class"""
        text = "%s" % NamingContext.__str__(self) +\
               "\n\tpartdn=%s" % self.partstr +\
               "".join("\n\tmsDS-NC-Replica-Locations=%s" % k for k in self.rw_location_list) +\
               "".join("\n\tmsDS-NC-RO-Replica-Locations=%s" % k for k in self.ro_location_list)
        return text


class Site(object):
    """An individual site object discovered thru the configuration
    naming context.  Contains all DSAs that exist within the site
    """
    def __init__(self, site_dnstr, nt_now):
        self.site_dnstr = site_dnstr
        self.site_guid = None
        self.site_options = 0
        self.site_topo_generator = None
        self.site_topo_failover = 0  # appears to be in minutes
        self.dsa_table = {}
        self.rw_dsa_table = {}
        self.nt_now = nt_now

    def load_site(self, samdb):
        """Loads the NTDS Site Settings options attribute for the site
        as well as querying and loading all DSAs that appear within
        the site.
        """
        ssdn = "CN=NTDS Site Settings,%s" % self.site_dnstr
        attrs = ["options",
                 "interSiteTopologyFailover",
                 "interSiteTopologyGenerator"]
        try:
            res = samdb.search(base=ssdn, scope=ldb.SCOPE_BASE,
                               attrs=attrs)
            self_res = samdb.search(base=self.site_dnstr, scope=ldb.SCOPE_BASE,
                                    attrs=['objectGUID'])
        except ldb.LdbError as e16:
            (enum, estr) = e16.args
            raise KCCError("Unable to find site settings for (%s) - (%s)" %
                           (ssdn, estr))

        msg = res[0]
        if "options" in msg:
            self.site_options = int(msg["options"][0])

        if "interSiteTopologyGenerator" in msg:
            self.site_topo_generator = \
                str(msg["interSiteTopologyGenerator"][0])

        if "interSiteTopologyFailover" in msg:
            self.site_topo_failover = int(msg["interSiteTopologyFailover"][0])

        msg = self_res[0]
        if "objectGUID" in msg:
            self.site_guid = misc.GUID(samdb.schema_format_value("objectGUID",
                                       msg["objectGUID"][0]))

        self.load_all_dsa(samdb)

    def load_all_dsa(self, samdb):
        """Discover all nTDSDSA thru the sites entry and
        instantiate and load the DSAs.  Each dsa is inserted
        into the dsa_table by dn string.
        """
        try:
            res = samdb.search(self.site_dnstr,
                               scope=ldb.SCOPE_SUBTREE,
                               expression="(objectClass=nTDSDSA)")
        except ldb.LdbError as e17:
            (enum, estr) = e17.args
            raise KCCError("Unable to find nTDSDSAs - (%s)" % estr)

        for msg in res:
            dnstr = str(msg.dn)

            # already loaded
            if dnstr in self.dsa_table:
                continue

            dsa = DirectoryServiceAgent(dnstr)

            dsa.load_dsa(samdb)

            # Assign this dsa to my dsa table
            # and index by dsa dn
            self.dsa_table[dnstr] = dsa
            if not dsa.is_ro():
                self.rw_dsa_table[dnstr] = dsa

    def get_dsa(self, dnstr):
        """Return a previously loaded DSA object by consulting
        the sites dsa_table for the provided DSA dn string

        :return: None if DSA doesn't exist
        """
        return self.dsa_table.get(dnstr)

    def select_istg(self, samdb, mydsa, ro):
        """Determine if my DC should be an intersite topology
        generator.  If my DC is the istg and is both a writeable
        DC and the database is opened in write mode then we perform
        an originating update to set the interSiteTopologyGenerator
        attribute in the NTDS Site Settings object.  An RODC always
        acts as an ISTG for itself.
        """
        # The KCC on an RODC always acts as an ISTG for itself
        if mydsa.dsa_is_ro:
            mydsa.dsa_is_istg = True
            self.site_topo_generator = mydsa.dsa_dnstr
            return True

        c_rep = get_dsa_config_rep(mydsa)

        # Load repsFrom and replUpToDateVector if not already loaded
        # so we can get the current state of the config replica and
        # whether we are getting updates from the istg
        c_rep.load_repsFrom(samdb)

        c_rep.load_replUpToDateVector(samdb)

        # From MS-ADTS 6.2.2.3.1 ISTG selection:
        #     First, the KCC on a writable DC determines whether it acts
        #     as an ISTG for its site
        #
        #     Let s be the object such that s!lDAPDisplayName = nTDSDSA
        #     and classSchema in s!objectClass.
        #
        #     Let D be the sequence of objects o in the site of the local
        #     DC such that o!objectCategory = s. D is sorted in ascending
        #     order by objectGUID.
        #
        # Which is a fancy way of saying "sort all the nTDSDSA objects
        # in the site by guid in ascending order".   Place sorted list
        # in D_sort[]
        D_sort = sorted(
            self.rw_dsa_table.values(),
            key=lambda dsa: ndr_pack(dsa.dsa_guid))

        # double word number of 100 nanosecond intervals since 1600s

        # Let f be the duration o!interSiteTopologyFailover seconds, or 2 hours
        # if o!interSiteTopologyFailover is 0 or has no value.
        #
        # Note: lastSuccess and ntnow are in 100 nanosecond intervals
        #       so it appears we have to turn f into the same interval
        #
        #       interSiteTopologyFailover (if set) appears to be in minutes
        #       so we'll need to convert to seconds and then 100 nanosecond
        #       intervals
        #       XXX [MS-ADTS] 6.2.2.3.1 says it is seconds, not minutes.
        #
        #       10,000,000 is number of 100 nanosecond intervals in a second
        if self.site_topo_failover == 0:
            f = 2 * 60 * 60 * 10000000
        else:
            f = self.site_topo_failover * 60 * 10000000

        # Let o be the site settings object for the site of the local
        # DC, or NULL if no such o exists.
        d_dsa = self.dsa_table.get(self.site_topo_generator)

        # From MS-ADTS 6.2.2.3.1 ISTG selection:
        #     If o != NULL and o!interSiteTopologyGenerator is not the
        #     nTDSDSA object for the local DC and
        #     o!interSiteTopologyGenerator is an element dj of sequence D:
        #
        if d_dsa is not None and d_dsa is not mydsa:
            # From MS-ADTS 6.2.2.3.1 ISTG Selection:
            #     Let c be the cursor in the replUpToDateVector variable
            #     associated with the NC replica of the config NC such
            #     that c.uuidDsa = dj!invocationId. If no such c exists
            #     (No evidence of replication from current ITSG):
            #         Let i = j.
            #         Let t = 0.
            #
            #     Else if the current time < c.timeLastSyncSuccess - f
            #     (Evidence of time sync problem on current ISTG):
            #         Let i = 0.
            #         Let t = 0.
            #
            #     Else (Evidence of replication from current ITSG):
            #         Let i = j.
            #         Let t = c.timeLastSyncSuccess.
            #
            # last_success appears to be a double word containing
            #     number of 100 nanosecond intervals since the 1600s
            j_idx = D_sort.index(d_dsa)

            found = False
            for cursor in c_rep.rep_replUpToDateVector_cursors:
                if d_dsa.dsa_ivid == cursor.source_dsa_invocation_id:
                    found = True
                    break

            if not found:
                i_idx = j_idx
                t_time = 0

            # XXX doc says current time < c.timeLastSyncSuccess - f
            # which is true only if f is negative or clocks are wrong.
            # f is not negative in the default case (2 hours).
            elif self.nt_now - cursor.last_sync_success > f:
                i_idx = 0
                t_time = 0
            else:
                i_idx = j_idx
                t_time = cursor.last_sync_success

        # Otherwise (Nominate local DC as ISTG):
        #     Let i be the integer such that di is the nTDSDSA
        #         object for the local DC.
        #     Let t = the current time.
        else:
            i_idx = D_sort.index(mydsa)
            t_time = self.nt_now

        # Compute a function that maintains the current ISTG if
        # it is alive, cycles through other candidates if not.
        #
        # Let k be the integer (i + ((current time - t) /
        #     o!interSiteTopologyFailover)) MOD |D|.
        #
        # Note: We don't want to divide by zero here so they must
        #       have meant "f" instead of "o!interSiteTopologyFailover"
        k_idx = (i_idx + ((self.nt_now - t_time) // f)) % len(D_sort)

        # The local writable DC acts as an ISTG for its site if and
        # only if dk is the nTDSDSA object for the local DC. If the
        # local DC does not act as an ISTG, the KCC skips the
        # remainder of this task.
        d_dsa = D_sort[k_idx]
        d_dsa.dsa_is_istg = True

        # Update if we are the ISTG, otherwise return
        if d_dsa is not mydsa:
            return False

        # Nothing to do
        if self.site_topo_generator == mydsa.dsa_dnstr:
            return True

        self.site_topo_generator = mydsa.dsa_dnstr

        # If readonly database then do not perform a
        # persistent update
        if ro:
            return True

        # Perform update to the samdb
        ssdn = "CN=NTDS Site Settings,%s" % self.site_dnstr

        m = ldb.Message()
        m.dn = ldb.Dn(samdb, ssdn)

        m["interSiteTopologyGenerator"] = \
            ldb.MessageElement(mydsa.dsa_dnstr, ldb.FLAG_MOD_REPLACE,
                               "interSiteTopologyGenerator")
        try:
            samdb.modify(m)

        except ldb.LdbError as estr:
            raise KCCError(
                "Could not set interSiteTopologyGenerator for (%s) - (%s)" %
                (ssdn, estr))
        return True

    def is_intrasite_topology_disabled(self):
        """Returns True if intra-site topology is disabled for site"""
        return (self.site_options &
                dsdb.DS_NTDSSETTINGS_OPT_IS_AUTO_TOPOLOGY_DISABLED) != 0

    def is_intersite_topology_disabled(self):
        """Returns True if inter-site topology is disabled for site"""
        return ((self.site_options &
                 dsdb.DS_NTDSSETTINGS_OPT_IS_INTER_SITE_AUTO_TOPOLOGY_DISABLED)
                != 0)

    def is_random_bridgehead_disabled(self):
        """Returns True if selection of random bridgehead is disabled"""
        return (self.site_options &
                dsdb.DS_NTDSSETTINGS_OPT_IS_RAND_BH_SELECTION_DISABLED) != 0

    def is_detect_stale_disabled(self):
        """Returns True if detect stale is disabled for site"""
        return (self.site_options &
                dsdb.DS_NTDSSETTINGS_OPT_IS_TOPL_DETECT_STALE_DISABLED) != 0

    def is_cleanup_ntdsconn_disabled(self):
        """Returns True if NTDS Connection cleanup is disabled for site"""
        return (self.site_options &
                dsdb.DS_NTDSSETTINGS_OPT_IS_TOPL_CLEANUP_DISABLED) != 0

    def same_site(self, dsa):
        """Return True if dsa is in this site"""
        if self.get_dsa(dsa.dsa_dnstr):
            return True
        return False

    def is_rodc_site(self):
        if len(self.dsa_table) > 0 and len(self.rw_dsa_table) == 0:
            return True
        return False

    def __str__(self):
        """Debug dump string output of class"""
        text = "%s:" % self.__class__.__name__ +\
               "\n\tdn=%s" % self.site_dnstr +\
               "\n\toptions=0x%X" % self.site_options +\
               "\n\ttopo_generator=%s" % self.site_topo_generator +\
               "\n\ttopo_failover=%d" % self.site_topo_failover
        for key, dsa in self.dsa_table.items():
            text = text + "\n%s" % dsa
        return text


class GraphNode(object):
    """A graph node describing a set of edges that should be directed to it.

    Each edge is a connection for a particular naming context replica directed
    from another node in the forest to this node.
    """

    def __init__(self, dsa_dnstr, max_node_edges):
        """Instantiate the graph node according to a DSA dn string

        :param max_node_edges: maximum number of edges that should ever
            be directed to the node
        """
        self.max_edges = max_node_edges
        self.dsa_dnstr = dsa_dnstr
        self.edge_from = []

    def __str__(self):
        text = "%s:" % self.__class__.__name__ +\
               "\n\tdsa_dnstr=%s" % self.dsa_dnstr +\
               "\n\tmax_edges=%d" % self.max_edges

        for i, edge in enumerate(self.edge_from):
            if isinstance(edge, str):
                text += "\n\tedge_from[%d]=%s" % (i, edge)

        return text

    def add_edge_from(self, from_dsa_dnstr):
        """Add an edge from the dsa to our graph nodes edge from list

        :param from_dsa_dnstr: the dsa that the edge emanates from
        """
        assert isinstance(from_dsa_dnstr, str)

        # No edges from myself to myself
        if from_dsa_dnstr == self.dsa_dnstr:
            return False
        # Only one edge from a particular node
        if from_dsa_dnstr in self.edge_from:
            return False
        # Not too many edges
        if len(self.edge_from) >= self.max_edges:
            return False
        self.edge_from.append(from_dsa_dnstr)
        return True

    def add_edges_from_connections(self, dsa):
        """For each nTDSConnection object associated with a particular
        DSA, we test if it implies an edge to this graph node (i.e.
        the "fromServer" attribute).  If it does then we add an
        edge from the server unless we are over the max edges for this
        graph node

        :param dsa: dsa with a dnstr equivalent to his graph node
        """
        for connect in dsa.connect_table.values():
            self.add_edge_from(connect.from_dnstr)

    def add_connections_from_edges(self, dsa, transport):
        """For each edge directed to this graph node, ensure there
           is a corresponding nTDSConnection object in the dsa.
        """
        for edge_dnstr in self.edge_from:
            connections = dsa.get_connection_by_from_dnstr(edge_dnstr)

            # For each edge directed to the NC replica that
            # "should be present" on the local DC, the KCC determines
            # whether an object c exists such that:
            #
            #    c is a child of the DC's nTDSDSA object.
            #    c.objectCategory = nTDSConnection
            #
            # Given the NC replica ri from which the edge is directed,
            #    c.fromServer is the dsname of the nTDSDSA object of
            #    the DC on which ri "is present".
            #
            #    c.options does not contain NTDSCONN_OPT_RODC_TOPOLOGY

            found_valid = False
            for connect in connections:
                if connect.is_rodc_topology():
                    continue
                found_valid = True

            if found_valid:
                continue

            # if no such object exists then the KCC adds an object
            # c with the following attributes

            # Generate a new dnstr for this nTDSConnection
            opt = dsdb.NTDSCONN_OPT_IS_GENERATED
            flags = (dsdb.SYSTEM_FLAG_CONFIG_ALLOW_RENAME |
                     dsdb.SYSTEM_FLAG_CONFIG_ALLOW_MOVE)

            dsa.new_connection(opt, flags, transport, edge_dnstr, None)

    def has_sufficient_edges(self):
        """Return True if we have met the maximum "from edges" criteria"""
        if len(self.edge_from) >= self.max_edges:
            return True
        return False


class Transport(object):
    """Class defines a Inter-site transport found under Sites
    """

    def __init__(self, dnstr):
        self.dnstr = dnstr
        self.options = 0
        self.guid = None
        self.name = None
        self.address_attr = None
        self.bridgehead_list = []

    def __str__(self):
        """Debug dump string output of Transport object"""

        text = "%s:\n\tdn=%s" % (self.__class__.__name__, self.dnstr) +\
               "\n\tguid=%s" % str(self.guid) +\
               "\n\toptions=%d" % self.options +\
               "\n\taddress_attr=%s" % self.address_attr +\
               "\n\tname=%s" % self.name +\
               "".join("\n\tbridgehead_list=%s" % dnstr for dnstr in self.bridgehead_list)

        return text

    def load_transport(self, samdb):
        """Given a Transport object with an prior initialization
        for the object's DN, search for the DN and load attributes
        from the samdb.
        """
        attrs = ["objectGUID",
                 "options",
                 "name",
                 "bridgeheadServerListBL",
                 "transportAddressAttribute"]
        try:
            res = samdb.search(base=self.dnstr, scope=ldb.SCOPE_BASE,
                               attrs=attrs)

        except ldb.LdbError as e18:
            (enum, estr) = e18.args
            raise KCCError("Unable to find Transport for (%s) - (%s)" %
                           (self.dnstr, estr))

        msg = res[0]
        self.guid = misc.GUID(samdb.schema_format_value("objectGUID",
                              msg["objectGUID"][0]))

        if "options" in msg:
            self.options = int(msg["options"][0])

        if "transportAddressAttribute" in msg:
            self.address_attr = str(msg["transportAddressAttribute"][0])

        if "name" in msg:
            self.name = str(msg["name"][0])

        if "bridgeheadServerListBL" in msg:
            for value in msg["bridgeheadServerListBL"]:
                dsdn = dsdb_dn_guess(samdb, value)
                dnstr = str(dsdn.dn)
                if dnstr not in self.bridgehead_list:
                    self.bridgehead_list.append(dnstr)


class RepsFromTo(object):
    """Class encapsulation of the NDR repsFromToBlob.

    Removes the necessity of external code having to
    understand about other_info or manipulation of
    update flags.
    """
    def __init__(self, nc_dnstr=None, ndr_blob=None):

        self.__dict__['to_be_deleted'] = False
        self.__dict__['nc_dnstr'] = nc_dnstr
        self.__dict__['update_flags'] = 0x0
        # XXX the following sounds dubious and/or better solved
        # elsewhere, but lets leave it for now. In particular, there
        # seems to be no reason for all the non-ndr generated
        # attributes to be handled in the round about way (e.g.
        # self.__dict__['to_be_deleted'] = False above). On the other
        # hand, it all seems to work. Hooray! Hands off!.
        #
        # WARNING:
        #
        # There is a very subtle bug here with python
        # and our NDR code.  If you assign directly to
        # a NDR produced struct (e.g. t_repsFrom.ctr.other_info)
        # then a proper python GC reference count is not
        # maintained.
        #
        # To work around this we maintain an internal
        # reference to "dns_name(x)" and "other_info" elements
        # of repsFromToBlob.  This internal reference
        # is hidden within this class but it is why you
        # see statements like this below:
        #
        #   self.__dict__['ndr_blob'].ctr.other_info = \
        #        self.__dict__['other_info'] = drsblobs.repsFromTo1OtherInfo()
        #
        # That would appear to be a redundant assignment but
        # it is necessary to hold a proper python GC reference
        # count.
        if ndr_blob is None:
            self.__dict__['ndr_blob'] = drsblobs.repsFromToBlob()
            self.__dict__['ndr_blob'].version = 0x1
            self.__dict__['dns_name1'] = None
            self.__dict__['dns_name2'] = None

            self.__dict__['ndr_blob'].ctr.other_info = \
                self.__dict__['other_info'] = drsblobs.repsFromTo1OtherInfo()

        else:
            self.__dict__['ndr_blob'] = ndr_blob
            self.__dict__['other_info'] = ndr_blob.ctr.other_info

            if ndr_blob.version == 0x1:
                self.__dict__['dns_name1'] = ndr_blob.ctr.other_info.dns_name
                self.__dict__['dns_name2'] = None
            else:
                self.__dict__['dns_name1'] = ndr_blob.ctr.other_info.dns_name1
                self.__dict__['dns_name2'] = ndr_blob.ctr.other_info.dns_name2

    def __str__(self):
        """Debug dump string output of class"""

        text = "%s:" % self.__class__.__name__ +\
               "\n\tdnstr=%s" % self.nc_dnstr +\
               "\n\tupdate_flags=0x%X" % self.update_flags +\
               "\n\tversion=%d" % self.version +\
               "\n\tsource_dsa_obj_guid=%s" % self.source_dsa_obj_guid +\
               ("\n\tsource_dsa_invocation_id=%s" %
                 self.source_dsa_invocation_id) +\
               "\n\ttransport_guid=%s" % self.transport_guid +\
               "\n\treplica_flags=0x%X" % self.replica_flags +\
               ("\n\tconsecutive_sync_failures=%d" %
                 self.consecutive_sync_failures) +\
               "\n\tlast_success=%s" % self.last_success +\
               "\n\tlast_attempt=%s" % self.last_attempt +\
               "\n\tdns_name1=%s" % self.dns_name1 +\
               "\n\tdns_name2=%s" % self.dns_name2 +\
               "\n\tschedule[ " +\
               "".join("0x%X " % slot for slot in self.schedule) +\
               "]"

        return text

    def __setattr__(self, item, value):
        """Set an attribute and change update flag.

        Be aware that setting any RepsFromTo attribute will set the
        drsuapi.DRSUAPI_DRS_UPDATE_ADDRESS update flag.
        """
        if item in ['schedule', 'replica_flags', 'transport_guid',
                    'source_dsa_obj_guid', 'source_dsa_invocation_id',
                    'consecutive_sync_failures', 'last_success',
                    'last_attempt']:

            if item in ['replica_flags']:
                self.__dict__['update_flags'] |= drsuapi.DRSUAPI_DRS_UPDATE_FLAGS
            elif item in ['schedule']:
                self.__dict__['update_flags'] |= drsuapi.DRSUAPI_DRS_UPDATE_SCHEDULE

            setattr(self.__dict__['ndr_blob'].ctr, item, value)

        elif item in ['dns_name1']:
            self.__dict__['dns_name1'] = value

            if self.__dict__['ndr_blob'].version == 0x1:
                self.__dict__['ndr_blob'].ctr.other_info.dns_name = \
                    self.__dict__['dns_name1']
            else:
                self.__dict__['ndr_blob'].ctr.other_info.dns_name1 = \
                    self.__dict__['dns_name1']

        elif item in ['dns_name2']:
            self.__dict__['dns_name2'] = value

            if self.__dict__['ndr_blob'].version == 0x1:
                raise AttributeError(item)
            else:
                self.__dict__['ndr_blob'].ctr.other_info.dns_name2 = \
                    self.__dict__['dns_name2']

        elif item in ['nc_dnstr']:
            self.__dict__['nc_dnstr'] = value

        elif item in ['to_be_deleted']:
            self.__dict__['to_be_deleted'] = value

        elif item in ['version']:
            raise AttributeError("Attempt to set readonly attribute %s" % item)
        else:
            raise AttributeError("Unknown attribute %s" % item)

        self.__dict__['update_flags'] |= drsuapi.DRSUAPI_DRS_UPDATE_ADDRESS

    def __getattr__(self, item):
        """Overload of RepsFromTo attribute retrieval.

        Allows external code to ignore substructures within the blob
        """
        if item in ['schedule', 'replica_flags', 'transport_guid',
                    'source_dsa_obj_guid', 'source_dsa_invocation_id',
                    'consecutive_sync_failures', 'last_success',
                    'last_attempt']:
            return getattr(self.__dict__['ndr_blob'].ctr, item)

        elif item in ['version']:
            return self.__dict__['ndr_blob'].version

        elif item in ['dns_name1']:
            if self.__dict__['ndr_blob'].version == 0x1:
                return self.__dict__['ndr_blob'].ctr.other_info.dns_name
            else:
                return self.__dict__['ndr_blob'].ctr.other_info.dns_name1

        elif item in ['dns_name2']:
            if self.__dict__['ndr_blob'].version == 0x1:
                raise AttributeError(item)
            else:
                return self.__dict__['ndr_blob'].ctr.other_info.dns_name2

        elif item in ['to_be_deleted']:
            return self.__dict__['to_be_deleted']

        elif item in ['nc_dnstr']:
            return self.__dict__['nc_dnstr']

        elif item in ['update_flags']:
            return self.__dict__['update_flags']

        raise AttributeError("Unknown attribute %s" % item)

    def is_modified(self):
        return (self.update_flags != 0x0)

    def set_unmodified(self):
        self.__dict__['update_flags'] = 0x0


class SiteLink(object):
    """Class defines a site link found under sites
    """

    def __init__(self, dnstr):
        self.dnstr = dnstr
        self.options = 0
        self.system_flags = 0
        self.cost = 0
        self.schedule = None
        self.interval = None
        self.site_list = []

    def __str__(self):
        """Debug dump string output of Transport object"""

        text = "%s:\n\tdn=%s" % (self.__class__.__name__, self.dnstr) +\
               "\n\toptions=%d" % self.options +\
               "\n\tsystem_flags=%d" % self.system_flags +\
               "\n\tcost=%d" % self.cost +\
               "\n\tinterval=%s" % self.interval

        if self.schedule is not None:
            text += "\n\tschedule.size=%s" % self.schedule.size +\
                    "\n\tschedule.bandwidth=%s" % self.schedule.bandwidth +\
                    ("\n\tschedule.numberOfSchedules=%s" %
                     self.schedule.numberOfSchedules)

            for i, header in enumerate(self.schedule.headerArray):
                text += ("\n\tschedule.headerArray[%d].type=%d" %
                         (i, header.type)) +\
                        ("\n\tschedule.headerArray[%d].offset=%d" %
                         (i, header.offset)) +\
                        "\n\tschedule.dataArray[%d].slots[ " % i +\
                        "".join("0x%X " % slot for slot in self.schedule.dataArray[i].slots) +\
                        "]"

        for guid, dn in self.site_list:
            text = text + "\n\tsite_list=%s (%s)" % (guid, dn)
        return text

    def load_sitelink(self, samdb):
        """Given a siteLink object with an prior initialization
        for the object's DN, search for the DN and load attributes
        from the samdb.
        """
        attrs = ["options",
                 "systemFlags",
                 "cost",
                 "schedule",
                 "replInterval",
                 "siteList"]
        try:
            res = samdb.search(base=self.dnstr, scope=ldb.SCOPE_BASE,
                               attrs=attrs, controls=['extended_dn:0'])

        except ldb.LdbError as e19:
            (enum, estr) = e19.args
            raise KCCError("Unable to find SiteLink for (%s) - (%s)" %
                           (self.dnstr, estr))

        msg = res[0]

        if "options" in msg:
            self.options = int(msg["options"][0])

        if "systemFlags" in msg:
            self.system_flags = int(msg["systemFlags"][0])

        if "cost" in msg:
            self.cost = int(msg["cost"][0])

        if "replInterval" in msg:
            self.interval = int(msg["replInterval"][0])

        if "siteList" in msg:
            for value in msg["siteList"]:
                dsdn = dsdb_dn_guess(samdb, value)
                guid = misc.GUID(dsdn.dn.get_extended_component('GUID'))
                dnstr = str(dsdn.dn)
                if (guid, dnstr) not in self.site_list:
                    self.site_list.append((guid, dnstr))

        if "schedule" in msg:
            self.schedule = ndr_unpack(drsblobs.schedule, value)
        else:
            self.schedule = new_connection_schedule()


class KCCFailedObject(object):
    def __init__(self, uuid, failure_count, time_first_failure,
                 last_result, dns_name):
        self.uuid = uuid
        self.failure_count = failure_count
        self.time_first_failure = time_first_failure
        self.last_result = last_result
        self.dns_name = dns_name


##################################################
# Global Functions and Variables
##################################################

def get_dsa_config_rep(dsa):
    # Find configuration NC replica for the DSA
    for c_rep in dsa.current_rep_table.values():
        if c_rep.is_config():
            return c_rep

    raise KCCError("Unable to find config NC replica for (%s)" %
                   dsa.dsa_dnstr)


def new_connection_schedule():
    """Create a default schedule for an NTDSConnection or Sitelink. This
    is packed differently from the repltimes schedule used elsewhere
    in KCC (where the 168 nibbles are packed into 84 bytes).
    """
    # 168 byte instances of the 0x01 value.  The low order 4 bits
    # of the byte equate to 15 minute intervals within a single hour.
    # There are 168 bytes because there are 168 hours in a full week
    # Effectively we are saying to perform replication at the end of
    # each hour of the week
    schedule = drsblobs.schedule()

    schedule.size = 188
    schedule.bandwidth = 0
    schedule.numberOfSchedules = 1

    header = drsblobs.scheduleHeader()
    header.type = 0
    header.offset = 20

    schedule.headerArray = [header]

    data = drsblobs.scheduleSlots()
    data.slots = [0x01] * 168

    schedule.dataArray = [data]
    return schedule


##################################################
# DNS related calls
##################################################

def uncovered_sites_to_cover(samdb, site_name):
    """
    Discover which sites have no DCs and whose lowest single-hop cost
    distance for any link attached to that site is linked to the site supplied.

    We compare the lowest cost of your single-hop link to this site to all of
    those available (if it exists). This means that a lower ranked siteLink
    with only the uncovered site can trump any available links (but this can
    only be done with specific, poorly enacted user configuration).

    If the site is connected to more than one other site with the same
    siteLink, only the largest site (failing that sorted alphabetically)
    creates the DNS records.

    :param samdb database
    :param site_name origin site (with a DC)

    :return a list of sites this site should be covering (for DNS)
    """
    sites_to_cover = []

    server_res = samdb.search(base=samdb.get_config_basedn(),
                              scope=ldb.SCOPE_SUBTREE,
                              expression="(&(objectClass=server)"
                              "(serverReference=*))")

    site_res = samdb.search(base=samdb.get_config_basedn(),
                            scope=ldb.SCOPE_SUBTREE,
                            expression="(objectClass=site)")

    sites_in_use = Counter()
    dc_count = 0

    # Assume server is of form DC,Servers,Site-ABCD because of schema
    for msg in server_res:
        site_dn = msg.dn.parent().parent()
        sites_in_use[site_dn.canonical_str()] += 1

        if site_dn.get_rdn_value().lower() == site_name.lower():
            dc_count += 1

    if len(sites_in_use) != len(site_res):
        # There is a possible uncovered site
        sites_uncovered = []

        for msg in site_res:
            if msg.dn.canonical_str() not in sites_in_use:
                sites_uncovered.append(msg)

        own_site_dn = "CN={},CN=Sites,{}".format(
            ldb.binary_encode(site_name),
            ldb.binary_encode(str(samdb.get_config_basedn()))
        )

        for site in sites_uncovered:
            encoded_dn = ldb.binary_encode(str(site.dn))

            # Get a sorted list of all siteLinks featuring the uncovered site
            link_res1 = samdb.search(base=samdb.get_config_basedn(),
                                     scope=ldb.SCOPE_SUBTREE, attrs=["cost"],
                                     expression="(&(objectClass=siteLink)"
                                     "(siteList={}))".format(encoded_dn),
                                     controls=["server_sort:1:0:cost"])

            # Get a sorted list of all siteLinks connecting this an the
            # uncovered site
            link_res2 = samdb.search(base=samdb.get_config_basedn(),
                                     scope=ldb.SCOPE_SUBTREE,
                                     attrs=["cost", "siteList"],
                                     expression="(&(objectClass=siteLink)"
                                     "(siteList={})(siteList={}))".format(
                                         own_site_dn,
                                         encoded_dn),
                                     controls=["server_sort:1:0:cost"])

            # Add to list if your link is equal in cost to lowest cost link
            if len(link_res1) > 0 and len(link_res2) > 0:
                cost1 = int(link_res1[0]['cost'][0])
                cost2 = int(link_res2[0]['cost'][0])

                # Own siteLink must match the lowest cost link
                if cost1 != cost2:
                    continue

                # In a siteLink with more than 2 sites attached, only pick the
                # largest site, and if there are multiple, the earliest
                # alphabetically.
                to_cover = True
                for site_val in link_res2[0]['siteList']:
                    site_dn = ldb.Dn(samdb, str(site_val))
                    site_dn_str = site_dn.canonical_str()
                    site_rdn = site_dn.get_rdn_value().lower()
                    if sites_in_use[site_dn_str] > dc_count:
                        to_cover = False
                        break
                    elif (sites_in_use[site_dn_str] == dc_count and
                          site_rdn < site_name.lower()):
                        to_cover = False
                        break

                if to_cover:
                    site_cover_rdn = site.dn.get_rdn_value()
                    sites_to_cover.append(site_cover_rdn.lower())

    return sites_to_cover



# Tests for the samba-tool user sub command reading Primary:userPassword
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from samba.tests.samba_tool.user_virtualCryptSHA_base import UserCmdCryptShaTestCase


class UserCmdCryptShaTestCaseUserPassword(UserCmdCryptShaTestCase):
    # gpg decryption not enabled.
    # both virtual attributes specified, no rounds option
    # no hashes stored in supplementalCredentials
    # Should not get values
    def test_no_gpg_both_hashes_no_rounds(self):
        self.add_user()
        out = self._get_password("virtualCryptSHA256,virtualCryptSHA512")

        self.assertTrue("virtualCryptSHA256:" not in out)
        self.assertTrue("virtualCryptSHA512:" not in out)
        self.assertTrue("rounds=" not in out)

    # gpg decryption not enabled.
    # SHA256 specified
    # no hashes stored in supplementalCredentials
    # No rounds
    #
    # Should not get values
    def test_no_gpg_sha256_no_rounds(self):
        self.add_user()
        out = self._get_password("virtualCryptSHA256")

        self.assertTrue("virtualCryptSHA256:" not in out)
        self.assertTrue("virtualCryptSHA512:" not in out)
        self.assertTrue("rounds=" not in out)

    # gpg decryption not enabled.
    # SHA512 specified
    # no hashes stored in supplementalCredentials
    # No rounds
    #
    # Should not get values
    def test_no_gpg_sha512_no_rounds(self):
        self.add_user()
        out = self._get_password("virtualCryptSHA512")

        self.assertTrue("virtualCryptSHA256:" not in out)
        self.assertTrue("virtualCryptSHA512:" not in out)
        self.assertTrue("rounds=" not in out)

    # gpg decryption not enabled.
    # SHA128 specified, i.e. invalid/unknown algorithm
    # no hashes stored in supplementalCredentials
    # No rounds
    #
    # Should not get values
    def test_no_gpg_invalid_alg_no_rounds(self):
        self.add_user()
        out = self._get_password("virtualCryptSHA128")

        self.assertTrue("virtualCryptSHA256:" not in out)
        self.assertTrue("virtualCryptSHA512:" not in out)
        self.assertTrue("rounds=" not in out)

    # gpg decryption not enabled.
    # both virtual attributes specified, no rounds option
    # both hashes stored in supplementalCredentials
    # Should get values
    def test_no_gpg_both_hashes_no_rounds_stored_hashes(self):
        self.add_user("CryptSHA512 CryptSHA256")

        out = self._get_password("virtualCryptSHA256,virtualCryptSHA512")

        self.assertTrue("virtualCryptSHA256:" in out)
        self.assertTrue("virtualCryptSHA512:" in out)
        self.assertTrue("rounds=" not in out)

        # Should be using the pre computed hash in supplementalCredentials
        # so it should not change between calls.
        sha256 = self._get_attribute(out, "virtualCryptSHA256")
        sha512 = self._get_attribute(out, "virtualCryptSHA512")

        out = self._get_password("virtualCryptSHA256,virtualCryptSHA512")
        self.assertEqual(sha256, self._get_attribute(out, "virtualCryptSHA256"))
        self.assertEqual(sha512, self._get_attribute(out, "virtualCryptSHA512"))

    # gpg decryption not enabled.
    # both virtual attributes specified, rounds specified
    # both hashes stored in supplementalCredentials, with not rounds
    # Should get hashes for the first matching scheme entry
    def test_no_gpg_both_hashes_rounds_stored_hashes(self):
        self.add_user("CryptSHA512 CryptSHA256")

        out = self._get_password("virtualCryptSHA256;rounds=2561," +
                                 "virtualCryptSHA512;rounds=5129")

        self.assertTrue("virtualCryptSHA256;rounds=2561:" in out)
        self.assertTrue("virtualCryptSHA512;rounds=5129:" in out)
        self.assertTrue("$rounds=" not in out)

        # Should be using the pre computed hash in supplementalCredentials
        # so it should not change between calls.
        sha256 = self._get_attribute(out, "virtualCryptSHA256;rounds=2561")
        sha512 = self._get_attribute(out, "virtualCryptSHA512;rounds=5129")

        out = self._get_password("virtualCryptSHA256,virtualCryptSHA512")
        self.assertEqual(sha256, self._get_attribute(out,
                                                     "virtualCryptSHA256"))
        self.assertEqual(sha512, self._get_attribute(out,
                                                     "virtualCryptSHA512"))

    # gpg decryption not enabled.
    # both virtual attributes specified, rounds specified
    # both hashes stored in supplementalCredentials, with rounds
    # Should get values
    def test_no_gpg_both_hashes_rounds_stored_hashes_with_rounds(self):
        self.add_user("CryptSHA512 " +
                      "CryptSHA256 " +
                      "CryptSHA512:rounds=5129 " +
                      "CryptSHA256:rounds=2561")

        out = self._get_password("virtualCryptSHA256;rounds=2561," +
                                 "virtualCryptSHA512;rounds=5129")

        self.assertTrue("virtualCryptSHA256;rounds=2561:" in out)
        self.assertTrue("virtualCryptSHA512;rounds=5129:" in out)
        self.assertTrue("$rounds=" in out)

        # Should be using the pre computed hash in supplementalCredentials
        # so it should not change between calls.
        sha256 = self._get_attribute(out, "virtualCryptSHA256;rounds=2561")
        sha512 = self._get_attribute(out, "virtualCryptSHA512;rounds=5129")

        out = self._get_password("virtualCryptSHA256;rounds=2561," +
                                 "virtualCryptSHA512;rounds=5129")
        self.assertEqual(sha256, self._get_attribute(out, "virtualCryptSHA256;rounds=2561"))
        self.assertEqual(sha512, self._get_attribute(out, "virtualCryptSHA512;rounds=5129"))

        # Number of rounds should match that specified
        self.assertTrue(sha256.startswith("{CRYPT}$5$rounds=2561"))
        self.assertTrue(sha512.startswith("{CRYPT}$6$rounds=5129"))

    # gpg decryption not enabled.
    # both virtual attributes specified, rounds specified
    # both hashes stored in supplementalCredentials, with rounds
    # number of rounds stored/requested do not match
    # Should get the precomputed hashes for CryptSHA512 and CryptSHA256
    def test_no_gpg_both_hashes_rounds_stored_hashes_with_rounds_no_match(self):
        self.add_user("CryptSHA512 " +
                      "CryptSHA256 " +
                      "CryptSHA512:rounds=5129 " +
                      "CryptSHA256:rounds=2561")

        out = self._get_password("virtualCryptSHA256;rounds=4000," +
                                 "virtualCryptSHA512;rounds=5000")

        self.assertTrue("virtualCryptSHA256;rounds=4000:" in out)
        self.assertTrue("virtualCryptSHA512;rounds=5000:" in out)
        self.assertTrue("$rounds=" not in out)

        # Should be using the pre computed hash in supplementalCredentials
        # so it should not change between calls.
        sha256 = self._get_attribute(out, "virtualCryptSHA256;rounds=4000")
        sha512 = self._get_attribute(out, "virtualCryptSHA512;rounds=5000")

        out = self._get_password("virtualCryptSHA256;rounds=4000," +
                                 "virtualCryptSHA512;rounds=5000")
        self.assertEqual(sha256, self._get_attribute(out, "virtualCryptSHA256;rounds=4000"))
        self.assertEqual(sha512, self._get_attribute(out, "virtualCryptSHA512;rounds=5000"))

        # As the number of rounds did not match, should have returned the
        # first hash of the corresponding scheme
        out = self._get_password("virtualCryptSHA256," +
                                 "virtualCryptSHA512")
        self.assertEqual(sha256, self._get_attribute(out, "virtualCryptSHA256"))
        self.assertEqual(sha512, self._get_attribute(out, "virtualCryptSHA512"))



# Samba4 AD database checker
#
# Copyright (C) Andrew Tridgell 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import optparse
import sys

import ldb
import samba.getopt as options
from samba import colour
from samba.auth import system_session
from samba.dbchecker import dbcheck
from samba.samdb import SamDB

from . import Command, CommandError, Option


class cmd_dbcheck(Command):
    """Check local AD database for errors."""
    synopsis = "%prog [<DN>] [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptionsDouble,
    }

    def process_yes(option, opt, value, parser):
        assert value is None
        rargs = parser.rargs
        if rargs:
            arg = rargs[0]
            if ((arg[:2] == "--" and len(arg) > 2) or
                (arg[:1] == "-" and len(arg) > 1 and arg[1] != "-")):
                setattr(parser.values, "yes", True)
            else:
                setattr(parser.values, "yes_rules", arg.split())
                del rargs[0]
        else:
            setattr(parser.values, "yes", True)

    takes_args = ["DN?"]

    takes_options = [
        Option("--scope", dest="scope", default="SUB",
               help="Pass search scope that builds DN list. Options: SUB, ONE, BASE"),
        Option("--fix", dest="fix", default=False, action='store_true',
               help='Fix any errors found'),
        Option("--yes", action='callback', callback=process_yes,
               help="don't confirm changes individually. Applies all as a single transaction (will not succeed if any errors are found)"),
        Option("--cross-ncs", dest="cross_ncs", default=False, action='store_true',
               help="cross naming context boundaries"),
        Option("-v", "--verbose", dest="verbose", action="store_true", default=False,
               help="Print more details of checking"),
        Option("-q", "--quiet", action="store_true", default=False,
               help="don't print details of checking"),
        Option("--attrs", dest="attrs", default=None, help="list of attributes to check (space separated)"),
        Option("--reindex", dest="reindex", default=False, action="store_true", help="force database re-index"),
        Option("--force-modules", dest="force_modules", default=False, action="store_true", help="force loading of Samba modules and ignore the @MODULES record (for very old databases)"),
        Option("--reset-well-known-acls",
               dest="reset_well_known_acls",
               default=False, action="store_true",
               help=("reset ACLs on objects with well known default values"
                     " (for updating from early 4.0.x)")),
        Option("--quick-membership-checks", dest="quick_membership_checks",
               help=("Skips missing/orphaned memberOf backlinks checks, "
                     "but speeds up dbcheck dramatically for domains with "
                     "large groups"),
               default=False, action="store_true"),
        Option("-H", "--URL", help="LDB URL for database or target server (defaults to local SAM database)",
               type=str, metavar="URL", dest="H"),
        Option("--selftest-check-expired-tombstones",
               dest="selftest_check_expired_tombstones", default=False, action="store_true",
               help=optparse.SUPPRESS_HELP),  # This is only used by tests
    ]

    def run(self, DN=None, H=None, verbose=False, fix=False, yes=False,
            cross_ncs=False, quiet=False,
            scope="SUB", credopts=None, sambaopts=None, versionopts=None,
            attrs=None, reindex=False, force_modules=False,
            quick_membership_checks=False,
            reset_well_known_acls=False,
            selftest_check_expired_tombstones=False,
            yes_rules=None):

        if yes_rules is None:
            yes_rules = []

        lp = sambaopts.get_loadparm()

        over_ldap = H is not None and H.startswith('ldap')

        if over_ldap:
            creds = credopts.get_credentials(lp, fallback_machine=True)
        else:
            creds = None

        if force_modules:
            samdb = SamDB(session_info=system_session(), url=H,
                          credentials=creds, lp=lp, options=["modules=samba_dsdb"])
        else:
            try:
                samdb = SamDB(session_info=system_session(), url=H,
                              credentials=creds, lp=lp)
            except:
                raise CommandError("Failed to connect to DB at %s.  If this is a really old sam.ldb (before alpha9), then try again with --force-modules" % H)

        if H is None or not over_ldap:
            samdb_schema = samdb
        else:
            samdb_schema = SamDB(session_info=system_session(), url=None,
                                 credentials=creds, lp=lp)

        scope_map = {"SUB": ldb.SCOPE_SUBTREE, "BASE": ldb.SCOPE_BASE, "ONE": ldb.SCOPE_ONELEVEL}
        scope = scope.upper()
        if scope not in scope_map:
            raise CommandError("Unknown scope %s" % scope)
        search_scope = scope_map[scope]

        controls = ['show_deleted:1']
        if over_ldap:
            controls.append('paged_results:1:1000')
        if cross_ncs:
            controls.append("search_options:1:2")

        if not attrs:
            attrs = ['*']
        else:
            attrs = attrs.split()

        # The dbcheck module always prints to stdout, not our self.outf
        # (yes, maybe FIXME).
        stdout_colour = colour.colour_if_wanted(sys.stdout,
                                                hint=self.requested_colour)

        started_transaction = False
        if yes and fix:
            samdb.transaction_start()
            started_transaction = True
        try:
            chk = dbcheck(samdb, samdb_schema=samdb_schema, verbose=verbose,
                          fix=fix, yes=yes, quiet=quiet,
                          in_transaction=started_transaction,
                          quick_membership_checks=quick_membership_checks,
                          reset_well_known_acls=reset_well_known_acls,
                          check_expired_tombstones=selftest_check_expired_tombstones,
                          colour=stdout_colour)

            for option in yes_rules:
                if hasattr(chk, option):
                    setattr(chk, option, 'ALL')
                else:
                    raise CommandError("Invalid fix rule %s" % option)

            if reindex:
                self.outf.write("Re-indexing...\n")
                error_count = 0
                if chk.reindex_database():
                    self.outf.write("completed re-index OK\n")

            elif force_modules:
                self.outf.write("Resetting @MODULES...\n")
                error_count = 0
                if chk.reset_modules():
                    self.outf.write("completed @MODULES reset OK\n")

            else:
                error_count = chk.check_database(DN=DN, scope=search_scope,
                                                 controls=controls, attrs=attrs)
        except:
            if started_transaction:
                samdb.transaction_cancel()
            raise

        if started_transaction:
            samdb.transaction_commit()

        if error_count != 0:
            sys.exit(1)



#
# Unix SMB/CIFS implementation.
# backend code for provisioning a Samba4 server

# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007-2008
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2008-2009
# Copyright (C) Oliver Liebel <oliver@itc.li> 2008-2009
#
# Based on the original in EJS:
# Copyright (C) Andrew Tridgell <tridge@samba.org> 2005
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Functions for setting up a Samba configuration (LDB and LDAP backends)."""

import shutil

class BackendResult(object):

    def report_logger(self, logger):
        """Rerport this result to a particular logger.

        """
        raise NotImplementedError(self.report_logger)


class ProvisionBackend(object):

    def __init__(self, paths=None, lp=None,
                 names=None, logger=None):
        """Provision a backend for samba4"""
        self.paths = paths
        self.lp = lp
        self.names = names
        self.logger = logger

        self.type = "ldb"

    def init(self):
        """Initialize the backend."""
        raise NotImplementedError(self.init)

    def start(self):
        """Start the backend."""
        raise NotImplementedError(self.start)

    def shutdown(self):
        """Shutdown the backend."""
        raise NotImplementedError(self.shutdown)

    def post_setup(self):
        """Post setup.

        :return: A BackendResult or None
        """
        raise NotImplementedError(self.post_setup)


class LDBBackend(ProvisionBackend):

    def init(self):

        # Wipe the old sam.ldb databases away
        shutil.rmtree(self.paths.samdb + ".d", True)

    def start(self):
        pass

    def shutdown(self):
        pass

    def post_setup(self):
        pass





# user management
#
# user setprimarygroup command
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba import ldb
from samba.auth import system_session
from samba.dcerpc import security
from samba.ndr import ndr_unpack
from samba.netcmd import Command, CommandError, Option
from samba.samdb import SamDB


class cmd_user_setprimarygroup(Command):
    """Set the primary group a user account.

This command sets the primary group a user account. The username specified on
the command is the sAMAccountName. The primarygroupname is the sAMAccountName
of the new primary group. The user must be a member of the group.

The command may be run from the root userid or another authorized userid. The
-H or --URL= option can be used to execute the command against a remote server.

Example1:
samba-tool user setprimarygroup TestUser1 newPrimaryGroup --URL=ldap://samba.samdom.example.com -Uadministrator%passw1rd

Example1 shows how to set the primary group for TestUser1 on a remote LDAP
server. The --URL parameter is used to specify the remote target server.  The
-U option is used to pass the username and password of a user that exists on
the remote server and is authorized to update the server.
"""
    synopsis = "%prog <username> <primarygroupname> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("-H", "--URL", help="LDB URL for database or target server", type=str,
               metavar="URL", dest="H"),
        ]

    takes_args = ["username", "primarygroupname"]

    def run(self, username, primarygroupname, credopts=None, sambaopts=None,
            versionopts=None, H=None):

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        samdb = SamDB(url=H, session_info=system_session(),
                      credentials=creds, lp=lp)

        filter = ("(&(sAMAccountName=%s)(objectClass=user))" %
                  ldb.binary_encode(username))
        try:
            res = samdb.search(base=samdb.domain_dn(),
                               expression=filter,
                               scope=ldb.SCOPE_SUBTREE,
                               controls=["extended_dn:1:1"],
                               attrs=["objectSid",
                                      "memberOf",
                                      "primaryGroupID"])
            user_sid_binary = res[0].get('objectSid', idx=0)
            user_sid = ndr_unpack(security.dom_sid, user_sid_binary)
            (user_dom_sid, user_rid) = user_sid.split()
            user_sid_dn = "<SID=%s>" % user_sid
            user_pgid = int(res[0].get('primaryGroupID', idx=0))
            user_groups = res[0].get('memberOf')
            if user_groups is None:
                user_groups = []
        except IndexError:
            raise CommandError("Unable to find user '%s'" % (username))

        user_group_sids = []
        for user_group in user_groups:
            user_group_dn = ldb.Dn(samdb, str(user_group))
            user_group_binary_sid = user_group_dn.get_extended_component("SID")
            user_group_sid = ndr_unpack(security.dom_sid, user_group_binary_sid)
            user_group_sids.append(user_group_sid)

        filter = ("(&(sAMAccountName=%s)(objectClass=group))" %
                  ldb.binary_encode(primarygroupname))
        try:
            res = samdb.search(base=samdb.domain_dn(),
                               expression=filter,
                               scope=ldb.SCOPE_SUBTREE,
                               attrs=["objectSid"])
            group_sid_binary = res[0].get('objectSid', idx=0)
        except IndexError:
            raise CommandError("Unable to find group '%s'" % (primarygroupname))

        primarygroup_sid = ndr_unpack(security.dom_sid, group_sid_binary)
        (primarygroup_dom_sid, primarygroup_rid) = primarygroup_sid.split()

        if user_dom_sid != primarygroup_dom_sid:
            raise CommandError("Group '%s' does not belong to the user's "
                               "domain" % primarygroupname)

        if primarygroup_rid != user_pgid and primarygroup_sid not in user_group_sids:
            raise CommandError("User '%s' is not member of group '%s'" %
                               (username, primarygroupname))

        setprimarygroup_ldif = """
dn: %s
changetype: modify
delete: primaryGroupID
primaryGroupID: %u
add: primaryGroupID
primaryGroupID: %u
""" % (user_sid_dn, user_pgid, primarygroup_rid)

        try:
            samdb.modify_ldif(setprimarygroup_ldif)
        except Exception as msg:
            raise CommandError("Failed to set primary group '%s' "
                               "for user '%s': %s" %
                               (primarygroupname, username, msg))
        self.outf.write("Changed primary group to '%s'\n" % primarygroupname)



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation.
#
# Copyright (C) Catalyst IT Ltd. 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

from enum import Enum, unique
import pyasn1

from samba.tests.krb5.kdc_base_test import KDCBaseTest
import samba.tests.krb5.rfc4120_pyasn1 as krb5_asn1
from samba.credentials import DONT_USE_KERBEROS
from samba.dcerpc import krb5pac
from samba.dcerpc.misc import SEC_CHAN_WKSTA
from samba.ndr import ndr_unpack
from samba.tests import DynamicTestCase
from samba.tests.krb5.rfc4120_constants import (
    AES256_CTS_HMAC_SHA1_96,
    AES128_CTS_HMAC_SHA1_96,
    ARCFOUR_HMAC_MD5,
    KDC_ERR_PREAUTH_REQUIRED,
    KRB_AS_REP,
    KU_AS_REP_ENC_PART,
    KRB_ERROR,
    KU_PA_ENC_TIMESTAMP,
    KU_TICKET,
    PADATA_ENC_TIMESTAMP,
    NT_ENTERPRISE_PRINCIPAL,
    NT_PRINCIPAL,
    NT_SRV_INST,
    KDC_ERR_C_PRINCIPAL_UNKNOWN,
)

global_asn1_print = False
global_hexdump = False


@unique
class TestOptions(Enum):
    Canonicalize = 1
    Enterprise = 2
    UpperRealm = 4
    UpperUserName = 8
    NetbiosRealm = 16
    UPN = 32
    RemoveDollar = 64
    AsReqSelf = 128
    Last = 256

    def is_set(self, x):
        return self.value & x


@unique
class CredentialsType(Enum):
    User = 1
    Machine = 2

    def is_set(self, x):
        return self.value & x


class TestData:

    def __init__(self, options, creds):
        self.options = options
        self.user_creds = creds
        self.user_name = self._get_username(options, creds)
        self.realm = self._get_realm(options, creds)

        if TestOptions.Enterprise.is_set(options):
            client_name_type = NT_ENTERPRISE_PRINCIPAL
        else:
            client_name_type = NT_PRINCIPAL

        self.cname = KDCBaseTest.PrincipalName_create(
            name_type=client_name_type, names=[self.user_name])
        if TestOptions.AsReqSelf.is_set(options):
            self.sname = self.cname
        else:
            self.sname = KDCBaseTest.PrincipalName_create(
                name_type=NT_SRV_INST, names=["krbtgt", self.realm])
        self.canonicalize = TestOptions.Canonicalize.is_set(options)

    def _get_realm(self, options, creds):
        realm = creds.get_realm()
        if TestOptions.NetbiosRealm.is_set(options):
            realm = creds.get_domain()
        if TestOptions.UpperRealm.is_set(options):
            realm = realm.upper()
        else:
            realm = realm.lower()
        return realm

    def _get_username(self, options, creds):
        name = creds.get_username()
        if TestOptions.RemoveDollar.is_set(options) and name.endswith("$"):
            name = name[:-1]
        if TestOptions.Enterprise.is_set(options):
            realm = creds.get_realm()
            name = "{0}@{1}".format(name, realm)
        if TestOptions.UpperUserName.is_set(options):
            name = name.upper()
        return name

    def __repr__(self):
        rep = "Test Data: "
        rep += "options = '" + "{:08b}".format(self.options) + "'"
        rep += ", user name = '" + self.user_name + "'"
        rep += ", realm = '" + self.realm + "'"
        rep += ", cname = '" + str(self.cname) + "'"
        rep += ", sname = '" + str(self.sname) + "'"
        return rep


MACHINE_NAME = "tstkrb5cnnmch"
USER_NAME = "tstkrb5cnnusr"


@DynamicTestCase
class KerberosASCanonicalizationTests(KDCBaseTest):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user_creds = None
        cls.machine_creds = None

    @classmethod
    def setUpDynamicTestCases(cls):

        def skip(ct, options):
            """ Filter out any mutually exclusive test options """
            if ct != CredentialsType.Machine and\
                    TestOptions.RemoveDollar.is_set(options):
                return True
            if ct != CredentialsType.Machine and\
                    TestOptions.AsReqSelf.is_set(options):
                return True
            return False

        def build_test_name(ct, options):
            name = "%sCredentials" % ct.name
            for opt in TestOptions:
                if opt.is_set(options):
                    name += ("_%s" % opt.name)
            return name

        for ct in CredentialsType:
            for x in range(TestOptions.Last.value):
                if skip(ct, x):
                    continue
                name = build_test_name(ct, x)
                cls.generate_dynamic_test("test", name, x, ct)

    def user_account_creds(self):
        if self.user_creds is None:
            samdb = self.get_samdb()
            type(self).user_creds, _ = self.create_account(samdb, USER_NAME)

        return self.user_creds

    def machine_account_creds(self):
        if self.machine_creds is None:
            samdb = self.get_samdb()
            type(self).machine_creds, _ = self.create_account(
                samdb,
                MACHINE_NAME,
                account_type=self.AccountType.COMPUTER)
            self.machine_creds.set_secure_channel_type(SEC_CHAN_WKSTA)
            self.machine_creds.set_kerberos_state(DONT_USE_KERBEROS)

        return self.machine_creds

    def setUp(self):
        super().setUp()
        self.do_asn1_print = global_asn1_print
        self.do_hexdump = global_hexdump

    def _test_with_args(self, x, ct):
        if ct == CredentialsType.User:
            creds = self.user_account_creds()
        elif ct == CredentialsType.Machine:
            creds = self.machine_account_creds()
        else:
            raise Exception("Unexpected credential type")
        data = TestData(x, creds)

        try:
            (rep, as_rep) = self.as_req(data)
        except pyasn1.error.PyAsn1Error as e:
            import traceback
            self.fail("ASN1 Error, Options {0:08b}:{1} {2}".format(
                data.options,
                traceback.format_exc(),
                e))
        # If as_req triggered an expected server error response
        # No need to test the response data.
        if rep is not None:
            # The kvno is optional, heimdal includes it
            # MIT does not.
            if 'kvno' in rep['enc-part']:
                kvno = rep['enc-part']['kvno']
                self.check_kvno(kvno, data)

            cname = rep['cname']
            self.check_cname(cname, data)

            crealm = rep['crealm'].decode('ascii')
            self.check_crealm(crealm, data)

            sname = as_rep['sname']
            self.check_sname(sname, data)

            srealm = as_rep['srealm'].decode('ascii')
            self.check_srealm(srealm, data)

            if TestOptions.AsReqSelf.is_set(data.options):
                ticket_creds = creds
            else:
                ticket_creds = self.get_krbtgt_creds()
            ticket_key = self.TicketDecryptionKey_from_creds(ticket_creds)

            ticket_encpart = rep['ticket']['enc-part']
            self.assertElementEqual(ticket_encpart, 'etype',
                                    ticket_key.etype)
            self.assertElementEqual(ticket_encpart, 'kvno',
                                    ticket_key.kvno)
            ticket_decpart = ticket_key.decrypt(KU_TICKET,
                                                ticket_encpart['cipher'])
            ticket_private = self.der_decode(
                ticket_decpart,
                asn1Spec=krb5_asn1.EncTicketPart())

            pac_data = self.get_pac(ticket_private['authorization-data'])
            pac = ndr_unpack(krb5pac.PAC_DATA, pac_data)

            for pac_buffer in pac.buffers:
                if pac_buffer.type == krb5pac.PAC_TYPE_LOGON_NAME:
                    if TestOptions.Canonicalize.is_set(data.options):
                        expected = data.user_creds.get_username()
                    else:
                        expected = data.user_name

                    self.assertEqual(expected, pac_buffer.info.account_name)
                    break
            else:
                self.fail('PAC_TYPE_LOGON_NAME not found')

    def as_req(self, data):
        user_creds = data.user_creds
        realm = data.realm

        cname = data.cname
        sname = data.sname

        till = self.get_KerberosTime(offset=36000)

        kdc_options = "0"
        if data.canonicalize:
            kdc_options = str(krb5_asn1.KDCOptions('canonicalize'))

        padata = None

        # Set the allowable encryption types
        etypes = (
            AES256_CTS_HMAC_SHA1_96,
            AES128_CTS_HMAC_SHA1_96,
            ARCFOUR_HMAC_MD5)

        req = self.AS_REQ_create(padata=padata,
                                 kdc_options=kdc_options,
                                 cname=cname,
                                 realm=realm,
                                 sname=sname,
                                 from_time=None,
                                 till_time=till,
                                 renew_time=None,
                                 nonce=0x7fffffff,
                                 etypes=etypes,
                                 addresses=None,
                                 additional_tickets=None)
        rep = self.send_recv_transaction(req)
        self.assertIsNotNone(rep)

        #
        # Check the protocol version, should be 5
        self.assertEqual(
            rep['pvno'], 5, "Data {0}".format(str(data)))

        self.assertEqual(
            rep['msg-type'], KRB_ERROR, "Data {0}".format(str(data)))

        if (not self.uncanonicalized_implicit_dollar and
            not data.canonicalize and
            TestOptions.RemoveDollar.is_set(data.options) and
            user_creds.get_username().endswith('$')):
            # We expect the principal not to be found because
            # a) smb.conf asked for foo -> foo$ to only match with canonicalization
            # b) we as client are not requesting canonicalization
            # c) we have removed a '$' from the true name
            #
            # These are the tests that combine "MachineCredentials"
            # with "RemoveDollar" but not "Canonicalize".
            self.check_error_rep(rep, KDC_ERR_C_PRINCIPAL_UNKNOWN)
            return (None, None)

        if self.require_canonicalization and not data.canonicalize:
            self.check_error_rep(rep, KDC_ERR_C_PRINCIPAL_UNKNOWN)
            # by returning None for the preauth rep, we avoid further
            # checks of the as_rep (second argument), which we don't
            # have.
            return (None, None)

        self.assertEqual(
            rep['error-code'],
            KDC_ERR_PREAUTH_REQUIRED,
            "Error code {0}, Data {1}".format(rep['error-code'], str(data)))

        rep_padata = self.der_decode(
            rep['e-data'], asn1Spec=krb5_asn1.METHOD_DATA())

        for pa in rep_padata:
            if pa['padata-type'] == 19:
                etype_info2 = pa['padata-value']
                break

        etype_info2 = self.der_decode(
            etype_info2, asn1Spec=krb5_asn1.ETYPE_INFO2())

        key = self.PasswordKey_from_etype_info2(user_creds, etype_info2[0])

        (patime, pausec) = self.get_KerberosTimeWithUsec()
        pa_ts = self.PA_ENC_TS_ENC_create(patime, pausec)
        pa_ts = self.der_encode(pa_ts, asn1Spec=krb5_asn1.PA_ENC_TS_ENC())

        pa_ts = self.EncryptedData_create(key, KU_PA_ENC_TIMESTAMP, pa_ts)
        pa_ts = self.der_encode(pa_ts, asn1Spec=krb5_asn1.EncryptedData())

        pa_ts = self.PA_DATA_create(PADATA_ENC_TIMESTAMP, pa_ts)

        kdc_options = "0"
        if data.canonicalize:
            kdc_options = str(krb5_asn1.KDCOptions('canonicalize'))
        padata = [pa_ts]

        req = self.AS_REQ_create(padata=padata,
                                 kdc_options=kdc_options,
                                 cname=cname,
                                 realm=realm,
                                 sname=sname,
                                 from_time=None,
                                 till_time=till,
                                 renew_time=None,
                                 nonce=0x7fffffff,
                                 etypes=etypes,
                                 addresses=None,
                                 additional_tickets=None)
        rep = self.send_recv_transaction(req)
        self.assertIsNotNone(rep)

        #
        # Check the protocol version, should be 5
        self.assertEqual(
            rep['pvno'], 5, "Data {0}".format(str(data)))

        msg_type = rep['msg-type']
        # Should not have got an error.
        # If we did, fail and print the error code to help debugging
        self.assertNotEqual(
            msg_type,
            KRB_ERROR,
            "Error code {0}, Data {1}".format(
                rep.get('error-code', ''),
                str(data)))

        self.assertEqual(msg_type, KRB_AS_REP, "Data {0}".format(str(data)))

        # Decrypt and decode the EncKdcRepPart
        enc = key.decrypt(KU_AS_REP_ENC_PART, rep['enc-part']['cipher'])
        if enc[0] == 0x7A:
            # MIT Kerberos Tags the EncASRepPart as a EncKDCRepPart
            # i.e. tag number 26 instead of tag number 25
            as_rep = self.der_decode(enc, asn1Spec=krb5_asn1.EncTGSRepPart())
        else:
            as_rep = self.der_decode(enc, asn1Spec=krb5_asn1.EncASRepPart())

        return (rep, as_rep)

    def check_cname(self, cname, data):
        if TestOptions.Canonicalize.is_set(data.options):
            expected_name_type = NT_PRINCIPAL
        elif TestOptions.Enterprise.is_set(data.options):
            expected_name_type = NT_ENTERPRISE_PRINCIPAL
        else:
            expected_name_type = NT_PRINCIPAL

        name_type = cname['name-type']
        self.assertEqual(
            expected_name_type,
            name_type,
            "cname name-type, Options {0:08b}".format(data.options))

        ns = cname['name-string']
        name = ns[0].decode('ascii')

        expected = data.user_name
        if TestOptions.Canonicalize.is_set(data.options):
            expected = data.user_creds.get_username()
        self.assertEqual(
            expected,
            name,
            "cname principal, Options {0:08b}".format(data.options))

    def check_crealm(self, crealm, data):
        realm = data.user_creds.get_realm()
        self.assertEqual(
            realm, crealm, "crealm, Options {0:08b}".format(data.options))

    def check_sname(self, sname, data):
        nt = sname['name-type']
        ns = sname['name-string']
        name = ns[0].decode('ascii')

        if TestOptions.AsReqSelf.is_set(data.options):
            expected_name_type = NT_PRINCIPAL
            if not TestOptions.Canonicalize.is_set(data.options)\
               and TestOptions.Enterprise.is_set(data.options):

                expected_name_type = NT_ENTERPRISE_PRINCIPAL

            self.assertEqual(
                expected_name_type,
                nt,
                "sname name-type, Options {0:08b}".format(data.options))
            expected = data.user_name
            if TestOptions.Canonicalize.is_set(data.options):
                expected = data.user_creds.get_username()
            self.assertEqual(
                expected,
                name,
                "sname principal, Options {0:08b}".format(data.options))
        else:
            self.assertEqual(
                NT_SRV_INST,
                nt,
                "sname name-type, Options {0:08b}".format(data.options))
            self.assertEqual(
                'krbtgt',
                name,
                "sname principal, Options {0:08b}".format(data.options))

            realm = ns[1].decode('ascii')
            expected = data.realm
            if TestOptions.Canonicalize.is_set(data.options):
                expected = data.user_creds.get_realm().upper()
            self.assertEqual(
                expected,
                realm,
                "sname realm, Options {0:08b}".format(data.options))

    def check_srealm(self, srealm, data):
        realm = data.user_creds.get_realm()
        self.assertEqual(
            realm, srealm, "srealm, Options {0:08b}".format(data.options))

    def check_kvno(self, kvno, data):
        self.assertEqual(
            1, kvno, "kvno, Options {0:08b}".format(data.options))


if __name__ == "__main__":
    global_asn1_print = False
    global_hexdump = False
    import unittest

    unittest.main()



# gp_firewalld_ext samba gpo policy
# Copyright (C) David Mulder <dmulder@suse.com> 2021
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from subprocess import Popen, PIPE
from shutil import which
import json
from samba.gp.gpclass import gp_pol_ext, gp_applier
from samba.gp.util.logging import log

def firewall_cmd(*args):
    fw_cmd = which('firewall-cmd')
    if fw_cmd is not None:
        cmd = [fw_cmd]
        cmd.extend(list(args))

        p = Popen(cmd, stdout=PIPE, stderr=PIPE)
        stdoutdata, _ = p.communicate()
        return p.returncode, stdoutdata
    else:
        return -1, 'firewall-cmd not found'

def rule_segment_parse(name, rule_segment):
    if isinstance(rule_segment, str):
        return ('%s=%s' % (name, rule_segment)) + ' '
    else:
        return '%s %s ' % (name,
            ' '.join(['%s=%s' % (k, v) for k, v in rule_segment.items()]))

class gp_firewalld_ext(gp_pol_ext, gp_applier):
    def __str__(self):
        return 'Security/Firewalld'

    def apply_zone(self, guid, zone):
        zone_attrs = []
        ret = firewall_cmd('--permanent', '--new-zone=%s' % zone)[0]
        if ret != 0:
            log.error('Failed to add new zone', zone)
        else:
            attribute = 'zone:%s' % zone
            self.cache_add_attribute(guid, attribute, zone)
            zone_attrs.append(attribute)
        # Default to matching the interface(s) for the default zone
        ret, out = firewall_cmd('--list-interfaces')
        if ret != 0:
            log.error('Failed to set interfaces for zone', zone)
        for interface in out.strip().split():
            ret = firewall_cmd('--permanent', '--zone=%s' % zone,
                               '--add-interface=%s' % interface.decode())
            if ret != 0:
                log.error('Failed to set interfaces for zone', zone)
        return zone_attrs

    def apply_rules(self, guid, rule_dict):
        rule_attrs = []
        for zone, rules in rule_dict.items():
            for rule in rules:
                if 'rule' in rule:
                    rule_parsed = rule_segment_parse('rule', rule['rule'])
                else:
                    rule_parsed = 'rule '
                for segment in ['source', 'destination', 'service', 'port',
                                'protocol', 'icmp-block', 'masquerade',
                                'icmp-type', 'forward-port', 'source-port',
                                'log', 'audit']:
                    names = [s for s in rule.keys() if s.startswith(segment)]
                    for name in names:
                        rule_parsed += rule_segment_parse(name, rule[name])
                actions = set(['accept', 'reject', 'drop', 'mark'])
                segments = set(rule.keys())
                action = actions.intersection(segments)
                if len(action) == 1:
                    rule_parsed += rule_segment_parse(list(action)[0],
                                                      rule[list(action)[0]])
                else:
                    log.error('Invalid firewall rule syntax')
                ret = firewall_cmd('--permanent', '--zone=%s' % zone,
                                   '--add-rich-rule', rule_parsed.strip())[0]
                if ret != 0:
                    log.error('Failed to add firewall rule', rule_parsed)
                else:
                    rhash = self.generate_value_hash(rule_parsed)
                    attribute = 'rule:%s:%s' % (zone, rhash)
                    self.cache_add_attribute(guid, attribute, rule_parsed)
                    rule_attrs.append(attribute)
        return rule_attrs

    def unapply(self, guid, attribute, value):
        if attribute.startswith('zone'):
            ret = firewall_cmd('--permanent',
                               '--delete-zone=%s' % value)[0]
            if ret != 0:
                log.error('Failed to remove zone', value)
            else:
                self.cache_remove_attribute(guid, attribute)
        elif attribute.startswith('rule'):
            _, zone, _ = attribute.split(':')
            ret = firewall_cmd('--permanent', '--zone=%s' % zone,
                               '--remove-rich-rule', value)[0]
            if ret != 0:
                log.error('Failed to remove firewall rule', value)
            else:
                self.cache_remove_attribute(guid, attribute)

    def apply(self, applier_func, *args):
        return applier_func(*args)

    def process_group_policy(self, deleted_gpo_list, changed_gpo_list):
        for guid, settings in deleted_gpo_list:
            if str(self) in settings:
                for attribute, value in settings[str(self)].items():
                    self.unapply(guid, attribute, value)

        for gpo in changed_gpo_list:
            if gpo.file_sys_path:
                section = 'Software\\Policies\\Samba\\Unix Settings\\Firewalld'
                pol_file = 'MACHINE/Registry.pol'
                path = os.path.join(gpo.file_sys_path, pol_file)
                pol_conf = self.parse(path)
                if not pol_conf:
                    continue
                attrs = []
                for e in pol_conf.entries:
                    if e.keyname.startswith(section):
                        if e.keyname.endswith('Rules'):
                            attrs.extend(self.apply(self.apply_rules, gpo.name,
                                                    json.loads(e.data)))
                        elif e.keyname.endswith('Zones'):
                            if e.valuename == '**delvals.':
                                continue
                            attrs.extend(self.apply(self.apply_zone, gpo.name,
                                                    e.data))

                # Cleanup all old zones and rules from this GPO
                self.clean(gpo.name, keep=attrs)

    def rsop(self, gpo):
        output = {}
        pol_file = 'MACHINE/Registry.pol'
        section = 'Software\\Policies\\Samba\\Unix Settings\\Firewalld'
        if gpo.file_sys_path:
            path = os.path.join(gpo.file_sys_path, pol_file)
            pol_conf = self.parse(path)
            if not pol_conf:
                return output
            for e in pol_conf.entries:
                if e.keyname.startswith(section):
                    if e.keyname.endswith('Zones'):
                        if e.valuename == '**delvals.':
                            continue
                        if 'Zones' not in output.keys():
                            output['Zones'] = []
                        output['Zones'].append(e.data)
                    elif e.keyname.endswith('Rules'):
                        if 'Rules' not in output.keys():
                            output['Rules'] = []
                        output['Rules'].append(json.loads(e.data))
        return output



# common functions for samba-tool python commands
#
# Copyright Andrew Tridgell 2010
# Copyright Giampaolo Lauria 2011 <lauria2@yahoo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
from samba.dcerpc import nbt
from samba.net import Net
from samba.netcmd import CommandError
import ldb


# In MS AD, setting a timeout to '(never)' corresponds to this value
NEVER_TIMESTAMP = int(-0x8000000000000000)


def _get_user_realm_domain(user, sam=None):
    r""" get the realm or the domain and the base user
        from user like:
        * username
        * DOMAIN\username
        * username@REALM

         A SamDB object can also be passed in to check
        our domain or realm against the obtained ones.
    """
    baseuser = user
    m = re.match(r"(\w+)\\(\w+$)", user)
    if m:
        domain = m.group(1)
        baseuser = m.group(2)

        if sam is not None:
            our_domain = sam.domain_netbios_name()
            if domain.lower() != our_domain.lower():
                raise CommandError(f"Given domain '{domain}' does not match "
                                   f"our domain '{our_domain}'!")

        return (baseuser.lower(), "", domain.upper())

    realm = ""
    m = re.match(r"(\w+)@(\w+)", user)
    if m:
        baseuser = m.group(1)
        realm = m.group(2)

        if sam is not None:
            our_realm = sam.domain_dns_name()
            our_realm_initial = our_realm.split('.', 1)[0]
            if realm.lower() != our_realm_initial.lower():
                raise CommandError(f"Given realm '{realm}' does not match our "
                                   f"realm '{our_realm}'!")

    return (baseuser.lower(), realm.upper(), "")


def netcmd_dnsname(lp):
    """return the full DNS name of our own host. Used as a default
       for hostname when running status queries"""
    return lp.get('netbios name').lower() + "." + lp.get('realm').lower()


def netcmd_finddc(lp, creds, realm=None):
    """Return domain-name of a writable/ldap-capable DC for the default
       domain (parameter "realm" in smb.conf) unless another realm has been
       specified as argument"""
    net = Net(creds=creds, lp=lp)
    if realm is None:
        realm = lp.get('realm')
    cldap_ret = net.finddc(domain=realm,
                           flags=nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_DS | nbt.NBT_SERVER_WRITABLE)
    return cldap_ret.pdc_dns_name


def netcmd_get_domain_infos_via_cldap(lp, creds, address=None):
    """Return domain information (CLDAP record) of the ldap-capable
       DC with the specified address"""
    net = Net(creds=creds, lp=lp)
    cldap_ret = net.finddc(address=address,
                           flags=nbt.NBT_SERVER_LDAP | nbt.NBT_SERVER_DS)
    return cldap_ret

def is_printable_attr_val(val):
    import unicodedata

    # The value must be convertible to a string value.
    try:
        str_val = str(val)
    except:
        return False

    # Characters of the Unicode Character Category "C" ("Other") are
    # supposed to be not printable. The category "C" includes control
    # characters, format specifier and others.
    for c in str_val:
        if unicodedata.category(c)[0] == 'C':
            return False

    return True

def get_ldif_for_editor(samdb, msg):

    # Copy the given message, because we do not
    # want to modify the original message.
    m = ldb.Message()
    m.dn = msg.dn

    for k in msg.keys():
        if k == "dn":
            continue
        vals = msg[k]
        m[k] = vals
        need_base64 = False
        for v in vals:
            if is_printable_attr_val(v):
                continue
            need_base64 = True
            break
        if not need_base64:
            m[k].set_flags(ldb.FLAG_FORCE_NO_BASE64_LDIF)

    result_ldif = samdb.write_ldif(m, ldb.CHANGETYPE_NONE)

    return result_ldif


def timestamp_to_mins(timestamp_str):
    """Converts a timestamp in -100 nanosecond units to minutes"""
    # treat a timestamp of 'never' the same as zero (this should work OK for
    # most settings, and it displays better than trying to convert
    # -0x8000000000000000 to minutes)
    if int(timestamp_str) == NEVER_TIMESTAMP:
        return 0
    else:
        return abs(int(timestamp_str)) / (1e7 * 60)


def timestamp_to_days(timestamp_str):
    """Converts a timestamp in -100 nanosecond units to days"""
    return timestamp_to_mins(timestamp_str) / (60 * 24)


def attr_default(msg, attrname, default):
    """get an attribute from a ldap msg with a default"""
    if attrname in msg:
        return msg[attrname][0]
    return default



# Unix SMB/CIFS implementation.
#
# Tests for `samba-tool user generate-csr`
#
# Copyright (C) Catalyst.Net Ltd 2025
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
from typing import Optional

from samba.domain.models import Computer, User
from samba.tests.samba_tool.base import SambaToolCmdTest

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


szOID_NTDS_CA_SECURITY_EXT = x509.ObjectIdentifier("1.3.6.1.4.1.311.25.2")


HOST = "ldap://{DC_SERVER}".format(**os.environ)
CREDS = "-U{DC_USERNAME}%{DC_PASSWORD}".format(**os.environ)


class SambaToolUserGenerateCsrTest(SambaToolCmdTest):
    cmd = "user"
    model = User
    user = "alice"

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.samdb = cls.getSamDB("-H", HOST, CREDS)

    def test_create_private_key_pem(self):
        self._test_create_private_key(encoding=serialization.Encoding.PEM)

    def test_create_private_key_der(self):
        self._test_create_private_key(encoding=serialization.Encoding.DER)

    def test_create_private_key_password(self):
        self._test_create_private_key(password="pass1234")

    def _test_create_private_key(
        self,
        *,
        encoding: serialization.Encoding = serialization.Encoding.PEM,
        password: Optional[str] = None,
    ):
        private_key = rsa.generate_private_key(
            public_exponent=65537, key_size=2048, backend=default_backend()
        )

        if password is not None:
            encryption_algorithm = serialization.BestAvailableEncryption(
                password.encode("utf-8")
            )
        else:
            encryption_algorithm = serialization.NoEncryption()

        private_key_bytes = private_key.private_bytes(
            encoding=encoding,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=encryption_algorithm,
        )

        user = self.model.find(self.samdb, self.user)
        account_name = user.account_name
        if user.user_principal_name is not None:
            account_upn = user.user_principal_name
        else:
            realm = self.samdb.domain_dns_name()
            account_upn = f"{account_name}@{realm.lower()}"

        subject_name = x509.Name([
            # Note that the subject name is used in certificate mappings
            x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "SambaState"),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "SambaSelfTesting"),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, "Users"),
            x509.NameAttribute(NameOID.COMMON_NAME, account_upn),
        ])
        subject_name_str = subject_name.rfc4514_string()

        with self.mktemp() as private_key_filename, self.mktemp() as csr_filename:
            with open(private_key_filename, "wb") as private_key_file:
                _ = private_key_file.write(private_key_bytes)

            cmd = [
                self.cmd,
                "generate-csr",
                "-H",
                HOST,
                CREDS,
                self.user,
                subject_name_str,
                private_key_filename,
                csr_filename,
            ]
            if password is not None:
                cmd.append("--private-key-pass")
                cmd.append(password)

            result, out, err = self.runcmd(*cmd)
            self.assertCmdSuccess(result, out, err)

            with open(csr_filename, "rb") as csr_file:
                csr_bytes = csr_file.read()

            csr = x509.load_pem_x509_csr(csr_bytes)

            self.assertEqual(subject_name, csr.subject)

            try:
                sid_extension = csr.extensions.get_extension_for_oid(
                    szOID_NTDS_CA_SECURITY_EXT,
                )
            except x509.ExtensionNotFound:
                self.fail("expected to find SID extension")

            # We don’t check the whole ASN.1 structure is correct, just that it
            # contains the encoded SID.
            encoded_sid = user.object_sid.encode("utf-8")
            self.assertIn(encoded_sid, sid_extension.value.value)


class SambaToolComputerGenerateCsrTest(SambaToolUserGenerateCsrTest):
    cmd = "computer"
    model = Computer
    user = "ADDC"


if __name__ == "__main__":
    global_asn1_print = False
    global_hexdump = False
    import unittest

    unittest.main()



# Unix SMB/CIFS implementation.
#
# claim type management
#
# Copyright (C) Catalyst.Net Ltd. 2023
#
# Written by Rob van der Linde <rob@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba.domain.models import AttributeSchema, ClaimType, ClassSchema
from samba.domain.models.exceptions import ModelError
from samba.netcmd import Command, CommandError, Option, SuperCommand


class cmd_domain_claim_claim_type_create(Command):
    """Create claim types on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--attribute", help="Attribute of claim type to create (required).",
               dest="attribute_name", action="store", type=str, required=True),
        Option("--class", help="Object classes to set claim type to.",
               dest="class_names", action="append", type=str, required=True),
        Option("--name", help="Optional display name or use attribute name.",
               dest="name", action="store", type=str),
        Option("--description",
               help="Optional description or use from attribute.",
               dest="description", action="store", type=str),
        Option("--disable", help="Disable claim type.",
               dest="disable", action="store_true"),
        Option("--enable", help="Enable claim type.",
               dest="enable", action="store_true"),
        Option("--protect",
               help="Protect claim type from accidental deletion.",
               dest="protect", action="store_true"),
        Option("--unprotect",
               help="Unprotect claim type from accidental deletion.",
               dest="unprotect", action="store_true")
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None,
            attribute_name=None, class_names=None, description=None,
            disable=None, enable=None, protect=None, unprotect=None):

        # mutually exclusive attributes
        if enable and disable:
            raise CommandError("--enable and --disable cannot be used together.")
        if protect and unprotect:
            raise CommandError("--protect and --unprotect cannot be used together.")

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        display_name = name or attribute_name
        try:
            claim_type = ClaimType.get(ldb, display_name=display_name)
        except ModelError as e:
            raise CommandError(e)

        # Check if a claim type with this display name already exists.
        # Note: you can register the same claim type under another display name.
        if claim_type:
            raise CommandError(f"Claim type {display_name} already exists, "
                               "but you can use --name to use another name.")

        # Either --enable will be set or --disable but never both.
        # The default if both are missing is enabled=True.
        if enable is not None:
            enabled = enable
        else:
            enabled = not disable

        # Lookup attribute and class names in schema.
        try:
            applies_to = [ClassSchema.find(ldb, name) for name in class_names]
            attribute = AttributeSchema.find(ldb, attribute_name)
            claim_type = ClaimType.new_claim_type(
                ldb, attribute, applies_to, display_name,
                description, enabled)
        except (ModelError, ValueError) as e:
            raise CommandError(e)

        # Create claim type
        try:
            claim_type.save(ldb)

            if protect:
                claim_type.protect(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Claim type created successfully.
        message = f"Created claim type: {display_name}"
        if attribute_name != display_name:
            message += f" ({attribute_name})"
        print(message, file=self.outf)


class cmd_domain_claim_claim_type_modify(Command):
    """Modify claim types on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Display name of claim type to modify (required).",
               dest="name", action="store", type=str, required=True),
        Option("--class", help="Object classes to set claim type to.",
               dest="class_names", action="append", type=str),
        Option("--description", help="Set the claim type description.",
               dest="description", action="store", type=str),
        Option("--enable",
               help="Enable claim type.",
               dest="enable", action="store_true"),
        Option("--disable",
               help="Disable claim type.",
               dest="disable", action="store_true"),
        Option("--protect",
               help="Protect claim type from accidental deletion.",
               dest="protect", action="store_true"),
        Option("--unprotect",
               help="Unprotect claim type from accidental deletion.",
               dest="unprotect", action="store_true")
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None,
            class_names=None, description=None, enable=None, disable=None,
            protect=None, unprotect=None):

        if enable and disable:
            raise CommandError("--enable and --disable cannot be used together.")
        if protect and unprotect:
            raise CommandError("--protect and --unprotect cannot be used together.")

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            claim_type = ClaimType.get(ldb, display_name=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if claim type exists.
        if not claim_type:
            raise CommandError(f"Claim type {name} not found.")

        # Either --enable will be set or --disable but never both.
        if enable:
            claim_type.enabled = True
        elif disable:
            claim_type.enabled = False

        # Update the description.
        if description is not None:
            claim_type.description = description

        # Change class names for claim type.
        if class_names is not None:
            try:
                applies_to = [ClassSchema.find(ldb, name) for name in class_names]
            except (ModelError, ValueError) as e:
                raise CommandError(e)

            claim_type.claim_type_applies_to_class = [obj.dn for obj in applies_to]

        # Update claim type.
        try:
            claim_type.save(ldb)

            if protect:
                claim_type.protect(ldb)
            elif unprotect:
                claim_type.unprotect(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Claim type updated successfully.
        print(f"Updated claim type: {name}", file=self.outf)


class cmd_domain_claim_claim_type_delete(Command):
    """Delete claim types on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Display name of claim type to delete (required).",
               dest="name", action="store", type=str, required=True),
        Option("--force", help="Force claim type delete even if it is protected.",
               dest="force", action="store_true")
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            name=None, force=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            claim_type = ClaimType.get(ldb, display_name=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if claim type exists first.
        if claim_type is None:
            raise CommandError(f"Claim type {name} not found.")

        # Delete claim type.
        try:
            if force:
                claim_type.unprotect(ldb)

            claim_type.delete(ldb)
        except ModelError as e:
            if not force:
                raise CommandError(
                    f"{e}\nTry --force to delete protected claim types.")

            raise CommandError(e)

        # Claim type deleted successfully.
        print(f"Deleted claim type: {name}", file=self.outf)


class cmd_domain_claim_claim_type_list(Command):
    """List claim types on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            output_format=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            claim_types = ClaimType.query(ldb)
        except ModelError as e:
            raise CommandError(e)

        # Using json output format gives more detail.
        if output_format == "json":
            self.print_json({claim_type.display_name: claim_type
                             for claim_type in claim_types})
        else:
            for claim_type in claim_types:
                print(claim_type.display_name, file=self.outf)


class cmd_domain_claim_claim_type_view(Command):
    """View a single claim type on the domain."""

    synopsis = "%prog -H <URL> [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Display name of claim type to view (required).",
               dest="name", action="store", type=str, required=True),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None):

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        try:
            claim_type = ClaimType.get(ldb, display_name=name)
        except ModelError as e:
            raise CommandError(e)

        # Check if claim type exists first.
        if claim_type is None:
            raise CommandError(f"Claim type {name} not found.")

        # Display claim type as JSON.
        self.print_json(claim_type.as_dict())


class cmd_domain_claim_claim_type(SuperCommand):
    """Manage claim types on the domain."""

    subcommands = {
        "create": cmd_domain_claim_claim_type_create(),
        "delete": cmd_domain_claim_claim_type_delete(),
        "modify": cmd_domain_claim_claim_type_modify(),
        "list": cmd_domain_claim_claim_type_list(),
        "view": cmd_domain_claim_claim_type_view(),
    }



# Tests to check the netlogon service is only running when it's required
#
# Copyright (C) Catalyst IT Ltd. 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from samba.tests import TestCase
import os

from samba.credentials import Credentials
from samba.dcerpc import netlogon
from samba import NTSTATUSError, ntstatus
import ctypes

"""
Tests whether the netlogon service is running
"""


class NetlogonServiceTests(TestCase):

    def setUp(self):
        super().setUp()

        self.server      = os.environ["SERVER"]
        self.lp          = self.get_loadparm()
        self.creds = Credentials()

        # prefer the DC user/password in environments that have it
        if "DC_USERNAME" in os.environ and "DC_PASSWORD" in os.environ:
            self.creds.set_username(os.environ["DC_USERNAME"])
            self.creds.set_password(os.environ["DC_PASSWORD"])
        else:
            self.creds.set_username(os.environ["USERNAME"])
            self.creds.set_password(os.environ["PASSWORD"])

        self.creds.guess(self.lp)

    def test_have_netlogon_connection(self):
        try:
            c = self.get_netlogon_connection()
            self.assertIsNotNone(c)
        except NTSTATUSError as e:
            # On non-DC test environments, netlogon should not be running on
            # the server, so we expect the test to fail here
            enum = ctypes.c_uint32(e.args[0]).value
            if enum == ntstatus.NT_STATUS_OBJECT_NAME_NOT_FOUND:
                self.fail("netlogon service is not running")
            else:
                raise

    # Establish netlogon connection over NP
    def get_netlogon_connection(self):
        return netlogon.netlogon("ncacn_np:%s[seal]" % self.server, self.lp,
                                 self.creds)
